#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  1 23:16:42 2025

@author: Samuel Nahmani (1,2)
https://www.ipgp.fr/annuaire/nahmani/)
contact : nahmani@ipgp.fr ou samuel.nahmani@ign.fr
(1) Université Paris Cité, Institut de physique du globe de Paris, CNRS, IGN, F-75005 Paris, France.
(2) Univ Gustave Eiffel, ENSG, IGN, F-77455 Marne-la-Vallée, France. 

Version: 1.0
Dépendances: pandas, numpy, geodezyx, datetime, gpsdatetime, gnsstoolbox

"""

#%%
# GeodeZYX Toolbox’s
# [Sakic et al., 2019]
# Sakic, Pierre; Mansur, Gustavo; Chaiyaporn, Kitpracha; Ballu, Valérie (2019): 
# The geodeZYX toolbox: a versatile Python 3 toolbox for geodetic-oriented purposes. v. 4.0.
# GFZ Data Services. http://doi.org/10.5880/GFZ.1.1.2019.002
#
# Documentation
# https://geodezyx.github.io/geodezyx-toolbox/
# 
# Installation
# pip install git+https://github.com/GeodeZYX/geodezyx-toolbox
# pip uninstall geodezyx

#%%
# gpsdatetime
# Python GPS date/time management package
# Copyright (C) 2014-2023, Jacques Beilin / ENSG-Geomatique
# Distributed under terms of the CECILL-C licence.
#%%
# GnssToolbox - Python package for GNSS learning
# Copyright (C) 2014-2023, Jacques Beilin / ENSG-Geomatique
# Distributed under terms of the CECILL-C licence.

#%%
# GeodeZYX Toolbox’s - [Sakic et al., 2019]
import geodezyx
import geodezyx.conv as conv                  # Import the conversion module
import datetime as dt

#
import gpsdatetime as gpst

import gnsstoolbox.orbits as orb
import gnsstoolbox.gnss_const as gnss_const
import gnsstoolbox.gnsstools as tools
import gnsstoolbox.gnss_process as proc
import gnsstoolbox.gnss_corr as gnss_corr

import pandas as pd
import numpy as np

# pour visualiser les données
import matplotlib.pyplot as plt

# mon module
from PPMD_mon_module import *

# pour chercher dans des dossiers
from pathlib import Path
import matplotlib.dates as mdates

#%%
# Chargement des fichiers RINEX d'observation
# <10 km de l'Etna : EMFN ESPC EINT ESLN ECPN EPLU ECNE DPDN
# <10-20 km du volcan : ECRI EMFN EBAG ESPC ESLN EMGL EMAL EDAM EMCN ECOR
# + EDEN station qui est loin de l'Etna


repertoire = Path("/home/snahmani/Traitements/IAG_WG4.3.9_IAVCEI_ETNA/ETNA_2018/GNSS_2018/rinex_2018")

# Filtrer et trier les fichiers
fichiers_sta = sorted(
    [f for f in repertoire.rglob("*ebag3610*.??o") if f.is_file() and f.stat().st_size > 0],
    key=lambda f: f.name
)

fichiers_sta = [str(f) for f in fichiers_sta]  # convertir en strings

# Lire les fichiers un par un
dfs = []
for fichier in fichiers_sta:
    print(f"Lecture de : {fichier}")
    try:
        df = geodezyx.files_rw.read_rinex2_obs(fichier, set_index=['epoch', 'prn'])  # <== [fichier] ici
        dfs.append(df)
    except Exception as e:
        print(f"Erreur lors de la lecture de {fichier} : {e}")
        
# Position approchée lue dans le header du fichier RINEX
columns =  grep_file(r"APPROX POSITION XYZ", fichier)
if columns:
    # Séparer la ligne en supprimant le texte "APPROX POSITION XYZ"
    valeurs = columns[0].split()[:3]  # Récupérer uniquement les 3 premières valeurs
    P_rnx_header = np.array(valeurs, dtype=float)  # Convertir en numpy array
    print("Coordonnées XYZ :", P_rnx_header)
else:
    print("Aucune correspondance trouvée.")

# Concaténer tous les DataFrames
if dfs:
    df_total = pd.concat(dfs)
    print("✅ Données combinées avec succès")
else:
    print("❌ Aucun fichier valide chargé")

#%%
# nettoyage
df_total = df_total.dropna(axis=1, how='all')

#%%
#
from datetime import datetime, timedelta
import os
import locale

# Définir la locale en anglais pour le formatage des dates
locale.setlocale(locale.LC_TIME, 'en_US.UTF-8')
# Extraire uniquement le nom du fichier (sans le chemin)
filename = os.path.basename(fichier)  # Résultat : 'kokb1430.01o'

try:
    # Extraction des informations DOY et année à partir du nom du fichier
    doy = int(filename[4:7])  # DOY (exemple : 143)
    year_suffix = filename[9:11]  # Suffixe d'année (exemple : '01')

    # Vérification que le suffixe d'année est bien numérique
    if not year_suffix.isdigit():
        raise ValueError(f"Invalid year suffix: {year_suffix}")

    # Conversion du suffixe de l'année en année complète
    if int(year_suffix) < 80:
        year = 2000 + int(year_suffix)
    else:
        year = 1900 + int(year_suffix)

    # Conversion du DOY en date complète
    date = datetime(year, 1, 1) + timedelta(days=doy - 1)

    # Formatage en date américaine
    date_american = date.strftime("%B %d, %Y")  # Exemple : "May 23, 2001"
    
    station = filename[0:4]

    # Créer le titre
    titre_generique = f"{station.upper()} - {date_american}"
    print(titre_generique)

except (ValueError, IndexError) as e:
    print(f"Error processing file '{filename}': {e}")


#%% Chargement des fichiers de navigation

repertoire = Path("/home/snahmani/Traitements/IAG_WG4.3.9_IAVCEI_ETNA/ETNA_2018/GNSS_2018/orb_2018")


# Filtrer et trier les fichiers
fichiers_sp3= sorted(
    [f for f in repertoire.rglob("igs*.sp3") if f.is_file() and f.stat().st_size > 0],
    key=lambda f: f.name
)

fichiers_sp3 = [str(f) for f in fichiers_sp3]  # convertir en strings

mysp3 = orb.orbit()
mysp3.loadSp3(fichiers_sp3)

#%%
#Il faut calculer la position de chaque satellite GNSS à chaque temps d'émission
t = gpst.gpsdatetime()

X_sat = []
Y_sat = []
Z_sat = []
dte_sat = []
dRelat = []

for (time_i,prn_i) in df_total.index:

    t.rinex_t(time_i.to_pydatetime().strftime('%y %m %d %H %M %S.%f'))
    
    t_emission_mjd  = t.mjd - df_total.loc[(time_i,prn_i), 'C1'] / gnss_const.c / 86400.0
    
    (X_sat_v,Y_sat_v,Z_sat_v,dte_sat_v)	 = mysp3.calcSatCoord(prn_i[0], int(prn_i[1:]),t_emission_mjd)
    
    # calcul de l'effet relativiste
    delta_t = 1e-3 # écart de temps en +/- pour calculer la dérivée 
    (Xs1,Ys1,Zs1,clocks1) = mysp3.calcSatCoord(prn_i[0], int(prn_i[1:]),t_emission_mjd - delta_t / 86400.0)    
    (Xs2,Ys2,Zs2,clocks2) = mysp3.calcSatCoord(prn_i[0], int(prn_i[1:]),t_emission_mjd + delta_t / 86400.0)  
    
    VX      = (np.array([Xs2-Xs1, Ys2-Ys1, Zs2-Zs1]))/2.0/delta_t
    VX0     = np.array([X_sat_v,Y_sat_v,Z_sat_v])

    dRelat_v  = -2.0 * VX0.T@ VX /(gnss_const.c **2)
    
    # temps d'emission du signal GNSS en temps GNSS (mjd)
    t_emission_mjd = t_emission_mjd - dte_sat_v / 86400.0 - dRelat_v / 86400.0
    
    # Recalcul de la position du satellite au temps d'emission (temps GNSS en mjd)   
    (X_sat_v,Y_sat_v,Z_sat_v,dte_sat_v)	 = mysp3.calcSatCoord(prn_i[0], int(prn_i[1:]),t_emission_mjd)
    
    
    X_sat.append(X_sat_v)
    Y_sat.append(Y_sat_v)
    Z_sat.append(Z_sat_v)
    dte_sat.append(dte_sat_v)
    dRelat.append(dRelat_v)
    
    
df_total['X_sat']   = X_sat 
df_total['Y_sat']   = Y_sat 
df_total['Z_sat']   = Z_sat
df_total['dte_sat'] = dte_sat
df_total['dRelat']  = dRelat


del X_sat, Y_sat, Z_sat, dte_sat, time_i, prn_i, t_emission_mjd, t, X_sat_v, Y_sat_v, Z_sat_v, dte_sat_v, dRelat_v, dRelat

#%%


Az_rad, Ele_rad = tools.toolAzEle(P_rnx_header[0],P_rnx_header[1],P_rnx_header[2],df_total.X_sat,df_total.Y_sat,df_total.Z_sat)
# radian to degree
rad2deg = 180 / np.pi

Ele_deg =  Ele_rad * rad2deg
Az_deg  =  Az_rad * rad2deg

# Ajout des colonnes au DataFrame
df_total['Az_deg'] = Az_deg
df_total['Ele_deg'] = Ele_deg

#%%
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

# Paramètres
gap_threshold = pd.Timedelta(minutes=15)

# Liste des PRNs uniques
prns = df_total.index.get_level_values('prn').unique()

# Création du plot
fig, ax = plt.subplots(figsize=(12, 6))

for prn in prns:
    # Extraire les données pour ce PRN
    data = df_total.xs(prn, level='prn').sort_index()
    
    # Récupérer les timestamps (epochs)
    times = data.index
    values = data['S2'].values

    # Calculer les deltas de temps
    time_deltas = times.to_series().diff()

    # Identifier les ruptures > 15 min
    rupture_idx = time_deltas > gap_threshold

    # On insère NaN juste après les ruptures
    times_with_nan = []
    values_with_nan = []

    for i in range(len(times)):
        times_with_nan.append(times[i])
        values_with_nan.append(values[i])
        
        # Si rupture après ce point → insérer NaN
        if i + 1 < len(times) and rupture_idx.iloc[i + 1]:
            times_with_nan.append(times[i] + pd.Timedelta(seconds=30))  # ajouter un point "vide"
            values_with_nan.append(np.nan)

    # Tracer la courbe
    ax.plot(times_with_nan, values_with_nan, label=prn)

# Mise en forme
ax.set_title("Séries temporelles S2 (L2) par PRN – avec coupures > 15 min")
ax.set_xlabel("Temps")
ax.set_ylabel("S2")
ax.legend(title="PRN")
#%%

import matplotlib.pyplot as plt
import numpy as np

# Extraction pour 'G10' sur une période spécifique avec df_index
start_period = pd.Timestamp('2018-12-27 00:00:00')
end_period = pd.Timestamp('2018-12-27 23:59:30')
extract2b = df_total.loc[(slice(start_period, end_period)), ['Az_deg', 'Ele_deg', 'S2']]

'Az_deg', 'Ele_deg',

# ⚠️ Filtrer les données valides
df_valid = extract2b

# Conversion en radians pour le tracé polaire
az = np.deg2rad(df_valid['Az_deg'].values)
ele = df_valid['Ele_deg'].values

# On convertit l’élévation pour que le zénith soit au centre
r = 90 - ele  # rayon = 0 au zénith, 90 à l’horizon

# Valeur à afficher (par exemple S2)
val = df_valid['S2'].values

# Création du skyplot
fig = plt.figure(figsize=(8, 8))
ax = fig.add_subplot(111, polar=True)

# Scatter polar avec couleurs selon la valeur
sc = ax.scatter(az, r, c=val, cmap='viridis', s=10, alpha=0.8)

# Personnalisation du plot
ax.set_theta_zero_location("N")  # 0° en haut
ax.set_theta_direction(-1)       # azimut dans le sens horaire
ax.set_rlim(0, 90)               # rayon de 0 (zénith) à 90 (horizon)
ax.set_title("Skyplot des satellites – coloration par S2", va='bottom')

# Barre de couleur
cbar = plt.colorbar(sc, ax=ax, orientation='vertical', pad=0.1)
cbar.set_label("S2")

plt.tight_layout()
plt.show()




#%%
fig = plt.figure(figsize=(8, 8))
ax = fig.add_subplot(111, polar=True)

sc = ax.scatter(az, r, c=val, cmap='viridis', s=10, alpha=0.8)

ax.set_theta_zero_location("N")
ax.set_theta_direction(-1)
ax.set_rlim(0, 90)
ax.set_title("Skyplot des satellites – coloration par S2", va='bottom', fontsize=12)

cbar = plt.colorbar(sc, ax=ax, orientation='vertical', pad=0.1)
cbar.set_label("S2")

# Ajouter une légende descriptive avec station + période
start_period = pd.Timestamp('2018-12-24 09:00:00')
end_period = pd.Timestamp('2018-12-24 17:00:00')
info_text = f"Station : EBAG\nPériode : {start_period.strftime('%Y-%m-%d %H:%M')} → {end_period.strftime('%H:%M')}"
fig.text(0.02, 0.98, info_text, fontsize=10, ha='left', va='top', transform=fig.transFigure)

plt.tight_layout()
plt.show()


#%%

import matplotlib.pyplot as plt
import numpy as np

# Extraction et préparation des données
df_reset = df_valid.reset_index()
df_reset_sorted = df_reset.sort_values('epoch')
first_points = df_reset_sorted.groupby('prn').first()

# Skyplot - coordonnées polaires
az = np.radians(df_reset['Az_deg'].values)
r = 90 - df_reset['Ele_deg'].values
val = df_reset['S2'].values

# Tracé du skyplot
fig = plt.figure(figsize=(8, 8))
ax = fig.add_subplot(111, polar=True)

sc = ax.scatter(az, r, c=val, cmap='viridis', s=10, alpha=0.8)

# Orientation classique : Nord en haut, rotation horaire
ax.set_theta_zero_location("N")
ax.set_theta_direction(-1)
ax.set_rlim(0, 90)
ax.set_title("Skyplot des satellites – coloration par S2", va='bottom', fontsize=12)

# Affichage du nom des satellites au début de leur arc
for prn, row in first_points.iterrows():
    az_rad = np.radians(row['Az_deg'])
    r_val = 90 - row['Ele_deg']
    ax.text(az_rad, r_val, prn, fontsize=8, fontweight='bold',
            ha='center', va='center', color='black', clip_on=True)

# Barre de couleur
cbar = plt.colorbar(sc, ax=ax, orientation='vertical', pad=0.1)
cbar.set_label("S2")

# Informations station/période
start_period = df_reset['epoch'].min()
end_period = df_reset['epoch'].max()
info_text = f"Station : EBAG\nPériode : {start_period.strftime('%Y-%m-%d %H:%M')} → {end_period.strftime('%H:%M')}"
fig.text(0.02, 0.98, info_text, fontsize=10, ha='left', va='top', transform=fig.transFigure)

plt.tight_layout()
plt.show()


#%%


import matplotlib.pyplot as plt
import numpy as np

# Extraction pour 'G10' sur une période spécifique avec df_index
start_period = pd.Timestamp('2018-12-27 16:00:00')
end_period = pd.Timestamp('2018-12-27 23:59:30')
extract2b = df_total.loc[(slice(start_period, end_period)), ['Az_deg', 'Ele_deg', 'S2']]

'Az_deg', 'Ele_deg',

# ⚠️ Filtrer les données valides
df_valid = extract2b

# Réinitialiser et trier le DataFrame
df_reset = df_valid.reset_index()
df_sorted = df_reset.sort_values('epoch')

# Récupérer le premier et le dernier point pour chaque PRN
first_points = df_sorted.groupby('prn').first()
last_points = df_sorted.groupby('prn').last()

# Skyplot - coordonnées polaires
az = np.radians(df_reset['Az_deg'].values)
r = 90 - df_reset['Ele_deg'].values
val = df_reset['S2'].values

# Tracé du skyplot
fig = plt.figure(figsize=(8, 8))
ax = fig.add_subplot(111, polar=True)

sc = ax.scatter(az, r, c=val, cmap='viridis', s=10, alpha=0.8)

# Orientation classique
ax.set_theta_zero_location("N")
ax.set_theta_direction(-1)
ax.set_rlim(0, 90)
ax.set_title("Skyplot des satellites – coloration par S2", va='bottom', fontsize=12)

# ➕ Étiquette au début de l’arc
for prn, row in first_points.iterrows():
    az_rad = np.radians(row['Az_deg'])
    r_val = 90 - row['Ele_deg']
    ax.text(az_rad, r_val, prn, fontsize=8, fontweight='bold',
            ha='center', va='center', color='black', clip_on=True)

# ➕ Étiquette à la fin de l’arc
for prn, row in last_points.iterrows():
    az_rad = np.radians(row['Az_deg'])
    r_val = 90 - row['Ele_deg']
    ax.text(az_rad, r_val, prn, fontsize=8, fontweight='bold',
            ha='center', va='center', color='black', clip_on=True)

# Barre de couleur
cbar = plt.colorbar(sc, ax=ax, orientation='vertical', pad=0.1)
cbar.set_label("S2")

# Informations station/période
start_period = df_reset['epoch'].min()
end_period = df_reset['epoch'].max()
info_text = f"Station : EBAG\nPériode : {start_period.strftime('%Y-%m-%d %H:%M')} → {end_period.strftime('%H:%M')}"
fig.text(0.02, 0.98, info_text, fontsize=10, ha='left', va='top', transform=fig.transFigure)

plt.tight_layout()
plt.show()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Author: Samuel Nahmani
# Date: 2025-08-14
# Description: Generate SPOTGINS tropospheric time series (.tropo) from listing files.
#              Reads gzip files in reverse order to keep only the last hourly estimation.
#              Produces an easy-to-read, pandas-friendly ISO 8601 date format.
#


import argparse
import gzip
import re
import datetime
from pathlib import Path


def parse_args():
    p = argparse.ArgumentParser(
        description="Create SPOTGINS tropospheric time series (.tropo) from listing files."
    )
    p.add_argument("-name", required=True, help="Station name (e.g., ABMF00GLP)")
    p.add_argument("-base_dir", required=True,
                   help="Base directory containing <STATION>/020_listings")
    p.add_argument("-output_dir", required=True,
                   help="Directory where the .tropo file will be written")
    p.add_argument("-stations_file",
                   default="/root/spotgins/metadata/stations/station_master_file.dat",
                   help="Path to station_master_file.dat (default: %(default)s)")
    return p.parse_args()


def read_master_file(station_name: str, master_file_path: Path):
    """Retourne un dict avec X,Y,Z,Lon,Lat,H ou None si fichier absent / station non trouvée."""
    if not master_file_path.is_file():
        print(f"Warning: master file not found at {master_file_path}")
        return None
    try:
        with master_file_path.open("r", encoding="utf-8", errors="ignore") as f:
            for line in f:
                # on cherche la station avec un peu de tolérance sur les espaces
                if f" {station_name} " in line or line.strip().startswith(station_name) or station_name in line:
                    fields = re.split(r"\s+", line.strip())
                    if len(fields) >= 9:
                        return {
                            "X": float(fields[6]),
                            "Y": float(fields[7]),
                            "Z": float(fields[8]),
                            "Lon": float(fields[3]),
                            "Lat": float(fields[4]),
                            "H": float(fields[5]),
                        }
    except Exception as e:
        print(f"Error reading master file: {e}")
    return None


def extract_data_from_gz(gz_path: Path):
    """
    Lit le .gz à l'envers, ne garde que la dernière estimation par horodatage.
    Retourne une liste triée: [iso_date, ZTD, ZTD_STD, ZHD, ZWD, TGN, STD_TGN, TGE, STD_TGE].
    """
    records_dict = {}
    try:
        with gzip.open(gz_path, "rt", encoding="utf-8", errors="ignore") as f:
            for line in f.readlines()[::-1]:  # lecture inversée
                if "SINEX_TT2:" not in line:
                    continue
                line = line.split("SINEX_TT2:", 1)[-1].strip()
                fields = re.split(r"\s+", line)
                # format attendu: [STATION, YY:DOY:SEC, ZTD, ZTD_STD, ZHD, ZWD, TGN, STD_TGN, TGE, STD_TGE]
                if len(fields) != 10:
                    continue
                try:
                    y, doy, sec = map(int, fields[1].split(":"))
                    y += 2000 if y < 100 else 0
                    dt = datetime.datetime.strptime(f"{y}-{doy:03d}", "%Y-%j") + datetime.timedelta(seconds=sec)
                    iso_date = dt.isoformat()
                except Exception:
                    continue

                # si on a déjà cet horodatage, on s'arrête pour ce fichier (on a la dernière estimation)
                if iso_date in records_dict:
                    break

                record = [
                    iso_date,
                    float(fields[2]),  # ZTD
                    float(fields[3]),  # ZTD_STD
                    float(fields[4]),  # ZHD
                    float(fields[5]),  # ZWD
                    float(fields[6]),  # TGN
                    float(fields[7]),  # STD_TGN
                    float(fields[8]),  # TGE
                    float(fields[9]),  # STD_TGE
                ]
                records_dict[iso_date] = record
    except Exception as e:
        print(f"Error reading {gz_path.name}: {e}")

    # tri par date ISO
    return list(sorted(records_dict.values(), key=lambda r: r[0]))


def generate_header(station_name: str, position: dict | None):
    """Construit l'entête SPOTGINS. Si position est None, utilise des 0.0."""
    if position is None:
        position = {"X": 0.0, "Y": 0.0, "Z": 0.0, "Lon": 0.0, "Lat": 0.0, "H": 0.0}
        center_note = "SPOTGINS (position unavailable)"
    else:
        center_note = "SPOTGINS - IPGP"

    now = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d at %H:%M:%S (UTC)")
    return (
        "# SPOTGINS SOLUTION [TROPO] v1\n"
        f"# Creation {now}\n"
        "#--------------------------------------\n"
        f"# STATION          : {station_name}\n"
        f"# ANALYSIS_CENTRE  : {center_note}\n"
        "# HOW_TO_CITE      : Santamaría-Gómez et al. (2025) https://doi.org/10.5194/essd-2025-223\n"
        "# STRATEGY_SUMMARY : https://en.poleterresolide.fr/geodesy-plotter-en/#/solution/SPOTGINS\n"
        "#--------------------------------------\n"
        "# REF_FRAME        : IGS20\n"
        "# PRODUCTS         : G20/GRG\n"
        "# CONSTELLATION    : GPS only until 2018-10, GPS+Galileo thereafter when available\n"
        "# UNITS            : meters\n"
        "# ELLIPSOID        : GRS80\n"
        "#--------------------------------------\n"
        f"# X_pos            : {position['X']:.6f}\n"
        f"# Y_pos            : {position['Y']:.6f}\n"
        f"# Z_pos            : {position['Z']:.6f}\n"
        f"# Longitude        : {position['Lon']:.6f}\n"
        f"# Latitude         : {position['Lat']:.6f}\n"
        f"# Height           : {position['H']:.6f}\n"
        "#--------------------------------------\n"
        "# Date                 ZTD        ZTD_STD    ZHD        ZWD        TGN        STD_TGN    TGE        STD_TGE\n"
        "# -------------------- ---------- ---------- ---------- ---------- ---------- ---------- ---------- ----------\n"
    )


def main():
    args = parse_args()
    station = args.name
    base_dir = Path(args.base_dir)
    output_dir = Path(args.output_dir)
    master_file = Path(args.stations_file)

    listings_dir = base_dir / station / "020_listings"
    output_file = output_dir / f"SPOTGINS_{station}.tropo"

    print(f"master file SPOTGINS : {master_file}")
    print(output_file)
    print(listings_dir)

    position = read_master_file(station, master_file)
    if not listings_dir.is_dir():
        print(f"Missing directory: {listings_dir}")
        return

    records = []
    for gz_file in sorted(listings_dir.glob("*.gz")):
        records.extend(extract_data_from_gz(gz_file))

    if not records:
        print(f"No valid records found for {station}.")
        return

    output_file.parent.mkdir(parents=True, exist_ok=True)
    with output_file.open("w", encoding="utf-8") as f:
        f.write(generate_header(station, position))
        for r in records:
            # largeur fixe, supporte les signes négatifs
            f.write(f"{r[0]:<20} " + " ".join(f"{v:>10.6f}" for v in r[1:]) + "\n")


if __name__ == "__main__":
    main()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 11 16:39:34 2025

@author: snahmani
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import statsmodels.api as sm
from scipy import stats

import plotly.graph_objects as go
import plotly.io as pio

# GeodeZYX Toolbox’s - [Sakic et al., 2019]
import geodezyx.files_rw as files_rw  # Import the file reading module
import geodezyx.conv as conv                  # Import the conversion module
import datetime as dt


import gnsstoolbox.orbits as orb
import gpsdatetime as gpst
import gnsstoolbox.gnss_const as gnss_const

from .klobuchar import *
import gnsstoolbox.gnsstools as tools

# Pour chercher des chaînes de caractère dans des fichiers
import re


def grep_file(pattern, filename):
    """Recherche un motif dans un fichier et retourne les lignes correspondantes."""
    results = []
    with open(filename, "r", encoding="utf-8") as file:
        for line in file:
            if re.search(pattern, line):
                results.append(line.strip())  # Supprime les espaces et les sauts de ligne

    return results  # Retourne les lignes trouvées

def get_approx_position(fichier):
    """
    Recherche la ligne contenant 'APPROX POSITION XYZ' dans le fichier
    et retourne les trois premières valeurs sous forme de numpy.array.
    """
    lignes = grep_file(r"APPROX POSITION XYZ", fichier)
    if lignes:
        # Récupérer la première ligne correspondante et les trois premières valeurs
        valeurs = lignes[0].split()[:3]
        return np.array(valeurs, dtype=float)
    else:
        return None


def load_and_clean_rinex(path):
    """
    Charge et nettoie un fichier RINEX d'observation.

    Paramètres :
      - path : chemin vers le fichier RINEX.

    Retourne :
      - df : DataFrame nettoyé, indexé par ['epoch', 'prn'],
             et contenant une colonne 'ind_ligne' indiquant le numéro de ligne.
    """
    # Chargement du fichier dans un DataFrame avec l'index ['epoch', 'prn']
    df = files_rw.read_rinex_obs(path, set_index=['epoch', 'prn'])

    # Suppression des colonnes entièrement vides
    df = df.dropna(axis=1, how='all')

    # Suppression des lignes manquant des mesures essentielles
    df = df.dropna(subset=['C1', 'L1', 'L2'])

    # Filtrage pour ne garder que les mesures des satellites GPS ('G')
    df = df[df['sys'].str.contains('G')]

    # On refait un dropna au cas où après filtrage certaines colonnes seraient entièrement vides
    df = df.dropna(axis=1, how='all')

    # Ajout d'une colonne d'indice de ligne basée sur la longueur du DataFrame
    df['ind_ligne'] = range(len(df))

    return df


def enrich_df_with_sat_positions(df, mysp3):
    """
    Pour chaque ligne du DataFrame df (avec un index composé de (epoch, prn)),
    calcule la position du satellite en prenant en compte le retard, l'effet
    relativiste et met à jour le DataFrame avec les colonnes X_sat, Y_sat, Z_sat,
    dte_sat et dRelat.

    Parameters:
      - df : pandas DataFrame dont l'index est (epoch, prn) et qui contient la colonne 'C1'
      - mysp3 : instance de l'objet d'orbite contenant la méthode calcSatCoord
      - t : instance de gpst.gpsdatetime() utilisée pour la conversion des temps

    Returns:
      - df enrichi avec les colonnes calculées.
    """

    # Listes pour stocker les résultats
    x_sat  = []
    y_sat  = []
    z_sat  = []
    dte_sat = []
    d_relat = []

    # Création de l'objet temps (pour la conversion)
    t = gpst.gpsdatetime()

    # Pour chaque observation du DataFrame
    for (time_i, prn_i) in df.index:
        # Conversion du temps d'observation au format attendu
        # Par exemple, on formate time_i en chaîne de caractère
        # t.rinex_t(time_i.to_pydatetime().strftime('%y %m %d %H %M %S.%f'))

        # Calcul du temps d'émission initial (mjd)
        t_emission_mjd = t.mjd - df.loc[(time_i, prn_i), 'C1'] / conv.SPEED_OF_LIGHT / 86400.0

        # Calcul initial de la position du satellite
        (x_sat_v, y_sat_v, z_sat_v, dte_sat_v) = mysp3.calcSatCoord(prn_i[0], int(prn_i[1:]), t_emission_mjd)

        # Calcul de l'effet relativiste à partir d'une dérivée numérique
        delta_t = 1e-3  # écart de temps en secondes
        (xs1, ys1, zs1, _) = mysp3.calcSatCoord(prn_i[0], int(prn_i[1:]), t_emission_mjd - delta_t/86400.0)
        (xs2, ys2, zs2, _) = mysp3.calcSatCoord(prn_i[0], int(prn_i[1:]), t_emission_mjd + delta_t/86400.0)

        # Estimation de la vitesse par différence centrée
        vx  = np.array([xs2 - xs1, ys2 - ys1, zs2 - zs1]) / (2.0 * delta_t)
        vx0 = np.array([x_sat_v, y_sat_v, z_sat_v])

        d_relat_v = -2.0 * vx0.T @ vx / (conv.SPEED_OF_LIGHT ** 2)

        # Correction du temps d'émission tenant compte du retard d'horloge et de l'effet relativiste
        t_emission_corr = t_emission_mjd - dte_sat_v / 86400.0 - d_relat_v / 86400.0

        # Recalcul de la position au temps corrigé
        (x_sat_v, y_sat_v, z_sat_v, dte_sat_v) = mysp3.calcSatCoord(prn_i[0], int(prn_i[1:]), t_emission_corr)

        # Stockage des résultats
        x_sat.append(x_sat_v)
        y_sat.append(y_sat_v)
        z_sat.append(z_sat_v)
        dte_sat.append(dte_sat_v)
        d_relat.append(d_relat_v)

    # Ajout des nouvelles colonnes au DataFrame
    df['X_sat']   = x_sat
    df['Y_sat']   = y_sat
    df['Z_sat']   = z_sat
    df['dte_sat'] = dte_sat
    df['dRelat']  = d_relat

    return df

def Sagnac_rotate_around_z(row):
    """
    Calcule la correction Sagnac pour les coordonnées satellites.
    Retourne un pd.Series avec les nouvelles colonnes X_sat_corr, Y_sat_corr, Z_sat_corr.
    """
    # Calcul de l'angle de rotation en radians
    alpha_rad = row['C1'] / conv.SPEED_OF_LIGHT * gnss_const.Omega_e
    # Construction de la matrice de rotation autour de l'axe Z
    Rz = np.array([[np.cos(alpha_rad), -np.sin(alpha_rad), 0],
                   [np.sin(alpha_rad),  np.cos(alpha_rad), 0],
                   [0, 0, 1]])
    # Vecteur de position original du satellite
    original_vector = np.array([row['X_sat'], row['Y_sat'], row['Z_sat']])
    # Calcul du vecteur corrigé
    rotated_vector = Rz.dot(original_vector)
    # Retourne les nouvelles coordonnées dans un pd.Series
    return pd.Series({
        'X_sat_sagnac': rotated_vector[0],
        'Y_sat_sagnac': rotated_vector[1],
        'Z_sat_sagnac': rotated_vector[2]
    })


def add_az_el_iono_columns(df, P_rnx_header, mynav):
    """
    Ajoute dans le DataFrame les colonnes Az (azimut en degrés), Ele (élévation en degrés)
    et d_ion1 (correction ionosphérique selon le modèle de Klobuchar).

    Paramètres :
      - df : DataFrame contenant les colonnes 'X_sat', 'Y_sat', 'Z_sat' et 'ind_ligne'.
             L'index doit être un MultiIndex (time, prn).
      - P_rnx_header : tableau ou liste contenant les coordonnées de la station (X, Y, Z)
      - mynav : objet de navigation contenant les attributs ion_alpha_gps et ion_beta_gps.
      - t : objet gpsdatetime (issu de gpst.gpsdatetime) pour la gestion des temps.
      - tools : module (par ex. gnsstoolbox.gnsstools) fournissant les fonctions toolAzEle et toolCartGeoGRS80.
      - klobuchar : module contenant la fonction klobuchar pour le calcul de la correction iono.

    Retourne :
      - df : le DataFrame enrichi avec les colonnes 'Az', 'Ele' et 'd_ion1'.
    """

    # Création de l'objet temps (pour la conversion)
    t = gpst.gpsdatetime()

    # Conversion des coordonnées de la station en géographiques
    lon, lat, h = conv.xyz2geo(P_rnx_header[0], P_rnx_header[1], P_rnx_header[2])
    rad2deg = 180 / np.pi
    lon_d = lon * rad2deg
    lat_d = lat * rad2deg

    # Calcul de l'azimut et de l'élévation en radians à partir des positions satellites
    # az_rad, ele_rad = tools.toolAzEle(P_rnx_header[0],
    #                                    P_rnx_header[1],
    #                                    P_rnx_header[2],
    #                                    df['X_sat'],
    #                                    df['Y_sat'],
    #                                    df['Z_sat'])

    az_rad , ele_rad = conv.xyz2azi_ele(df['X_sat'], df['Y_sat'], df['Z_sat'],
                                        P_rnx_header[0], P_rnx_header[1], P_rnx_header[2])

    # Conversion en degrés
    az_deg = az_rad * rad2deg
    ele_deg = ele_rad * rad2deg

    # Récupération des coefficients ionosphériques depuis le fichier de navigation
    alpha = mynav.ion_alpha_gps
    beta  = mynav.ion_beta_gps

    # Calcul de la correction iono pour chaque observation
    d_ion1 = []
    for (time_i, prn_i) in df.index:
        # Mise à jour de l'objet temps pour cette observation (conversion via rinex_t)
        #t.rinex_t(time_i.to_pydatetime().strftime('%y %m %d %H %M %S.%f'))
        #wsec_v = t.wsec  # secondes dans la semaine GPS

        _, wsec_v = conv.dt2gpstime(time_i, secinweek=True)

        # Récupération de l'indice de ligne correspondant (doit être un entier valide)
        i = df.loc[(time_i, prn_i), 'ind_ligne']

        # Calcul de la correction ionosphérique (la fonction klobuchar doit accepter ces paramètres)
        d_ion1_v = klobuchar.klobuchar(lat_d, lon_d, ele_deg[i], az_deg[i], wsec_v, alpha, beta)
        d_ion1.append(d_ion1_v)

    # Ajout des nouvelles colonnes dans le DataFrame
    df['Az']    = az_deg
    df['Ele']   = ele_deg
    df['d_ion1'] = d_ion1

    return df


def plot_series(df, col1, col2=None, coeff1=1.0, coeff2=1.0, seuil=3600, renderer="browser"):
    """
    Affiche les séries temporelles pour chaque satellite.
    Si col2 est fourni, affiche la série : coeff1 * col1 - coeff2 * col2.
    Sinon, affiche la série de la colonne col1 directement.
    La série est découpée en segments lorsqu'un "trou" (écart > seuil) est détecté.

    Paramètres :
        df      : DataFrame avec un index multi-niveaux contenant au moins le niveau 'prn'
                  et les colonnes col1 (et éventuellement col2).
        col1    : Nom de la première colonne.
        col2    : Nom de la seconde colonne (optionnel). Si None, on affiche col1.
        coeff1  : Coefficient multiplicateur pour la première colonne (défaut 1.0).
        coeff2  : Coefficient multiplicateur pour la seconde colonne (défaut 1.0).
        seuil   : Seuil en secondes pour considérer un "trou" dans la série (défaut 3600).
        renderer: Renderer Plotly (ex : "browser" ou "iframe").

    Retourne :
        fig     : Figure Plotly contenant les courbes tracées.
    """
    # Configuration du renderer
    pio.renderers.default = renderer

    # Extraction et tri de la liste des satellites (niveau 'prn')
    satellites = sorted(df.index.get_level_values('prn').unique())

    fig = go.Figure()

    for sat in satellites:
        # Sélection des données pour le satellite
        sub_df = df.xs(sat, level='prn')

        # Calcul de la série
        if col2 is not None:
            ts = (sub_df[col1] * coeff1 - sub_df[col2] * coeff2).dropna()
            y_label = f"{coeff1}*{col1} - {coeff2}*{col2}"
        else:
            ts = (sub_df[col1] * coeff1).dropna()
            y_label = f"{coeff1}*{col1}"

        dates = ts.index

        # Calcul des écarts entre dates successives (en secondes)
        ecarts = dates.to_series().diff().dt.total_seconds()

        # Identification des positions où l'écart dépasse le seuil
        breaks = np.where(ecarts > seuil)[0]

        # Découpage de la série en segments
        segments_dates = np.split(dates, breaks)
        segments_values = np.split(ts.values, breaks)

        # Reconstitution de la série en insérant des valeurs None pour créer des "trous"
        combined_dates = []
        combined_values = []
        for seg_dates, seg_values in zip(segments_dates, segments_values):
            if len(seg_dates) > 0:
                combined_dates.extend(seg_dates)
                combined_values.extend(seg_values)
                # Insertion d'un "trou" (None) pour forcer une cassure
                combined_dates.append(None)
                combined_values.append(None)
        # Supprimer le dernier trou ajouté si présent
        if combined_dates and combined_dates[-1] is None:
            combined_dates = combined_dates[:-1]
            combined_values = combined_values[:-1]

        # Ajout d'une trace unique par satellite
        fig.add_trace(go.Scatter(
            x=combined_dates,
            y=combined_values,
            mode='lines',
            name=f"Sat {sat}",
            legendgroup=f"sat_{sat}"
        ))

    # Boutons pour sélectionner/désélectionner les traces
    buttons = [
        dict(
            label="Tout sélectionner",
            method="update",
            args=[{"visible": [True] * len(fig.data)}]
        ),
        dict(
            label="Tout désélectionner",
            method="update",
            args=[{"visible": ["legendonly"] * len(fig.data)}]
        )
    ]

    # Mise à jour de la mise en page du graphique
    fig.update_layout(
        title=dict(
            text=y_label+" par satellite",
            x=0.5,
            xanchor="center"
        ),
        xaxis_title="Date",
        yaxis_title=y_label,
        hovermode="x unified",
        updatemenus=[dict(
            type="buttons",
            direction="left",
            buttons=buttons,
            pad={"r": 10, "t": 10},
            showactive=True,
            x=0.0,
            xanchor="left",
            y=1.1,   # Position des boutons en dehors de la zone de tracé, en haut
            yanchor="top"
        )],
        margin=dict(t=120)  # Augmentation de la marge supérieure pour que le titre et les boutons ne se chevauchent pas
    )

    fig.show()
    return fig


def plot_residual_analysis(A, B, dP_est, figure_title=None, save_path=None,
                           P_est=None, P_rnx_header=None, tools=None):
    """
    Calcule les résidus (v_est = B - A @ dP_est) et trace une figure contenant :
      1. La série temporelle des résidus (affichée en points)
      2. L'histogramme des résidus (nombre d'observations par bin)
      3. Le Q-Q Plot des résidus
      4. Un scatter plot des résidus en fonction des valeurs prédites
      5. Un encadré affichant quelques statistiques (moyenne, variance, écart-type, skewness, kurtosis)
      6. (Optionnel) Un encadré avec des informations supplémentaires sur la position :
         - Distance entre la position estimée et la position initiale du header RINEX
         - Coordonnées ENU locales calculées via tools.toolCartLocGRS80
         
    Paramètres :
      - A : array-like, matrice des variables explicatives.
      - B : array-like, vecteur des observations.
      - dP_est : array-like, vecteur des paramètres estimés.
      - figure_title (optionnel) : str, titre global de la figure.
      - save_path (optionnel) : str, chemin complet (nom + extension) pour sauvegarder la figure.
      - P_est (optionnel) : array-like, position estimée (pour le calcul des informations supplémentaires).
      - P_rnx_header (optionnel) : array-like, position initiale extraite du header RINEX.
      - tools (optionnel) : module ou objet possédant la fonction toolCartLocGRS80.
      
    Renvoie :
      - fig : l'objet Figure de matplotlib contenant l'ensemble des graphiques.
    """
    
    # Calcul des résidus et des valeurs prédites
    v_est = B - A @ dP_est
    b_est = A @ dP_est
    
    # Calcul des statistiques sur les résidus
    moyenne    = np.mean(v_est)
    variance   = np.var(v_est)
    ecart_type = np.std(v_est)
    skewness   = stats.skew(v_est)
    kurtosis   = stats.kurtosis(v_est)
    
    # Création de la figure avec GridSpec.
    # On utilise 4 lignes et 2 colonnes :
    # - Ligne 0 : Série temporelle (col 0) et Histogramme (col 1)
    # - Ligne 1 : Q-Q Plot (col 0) et Résidus vs. Valeurs prédites (col 1)
    # - Ligne 2 : Encadré des statistiques (s'étendant sur 2 colonnes)
    # - Ligne 3 : (Optionnel) Encadré des informations de position (s'étendant sur 2 colonnes)
    nrows = 4 if (P_est is not None and P_rnx_header is not None and tools is not None) else 3
    height_ratios = [1, 1, 0.5, 0.5] if nrows == 4 else [1, 1, 0.5]
    
    fig = plt.figure(figsize=(15, 15))
    gs = fig.add_gridspec(nrows, 2, height_ratios=height_ratios)
    
    # 1. Série temporelle des résidus (affichage uniquement des points) – Top gauche
    ax_time = fig.add_subplot(gs[0, 0])
    ax_time.scatter(np.arange(len(v_est)), v_est, color='green')
    ax_time.set_title("Série temporelle des résidus")
    ax_time.set_xlabel("Temps / Index")
    ax_time.set_ylabel("Résidus")
    
    # 2. Histogramme des résidus (nombre brut d'observations) – Top droite
    ax_hist = fig.add_subplot(gs[0, 1])
    sns.histplot(v_est, bins=30, stat="count", color='skyblue', edgecolor='black', ax=ax_hist)
    ax_hist.set_title("Histogramme des résidus")
    ax_hist.set_xlabel("Résidus")
    ax_hist.set_ylabel("Nombre d'observations")
    
    # 3. Q-Q Plot des résidus – Ligne 1, Colonne 0
    ax_qq = fig.add_subplot(gs[1, 0])
    sm.qqplot(v_est, line='s', ax=ax_qq)
    ax_qq.set_title("Q-Q Plot des résidus")
    
    # 4. Graphique des résidus vs. valeurs prédites – Ligne 1, Colonne 1
    ax_scatter = fig.add_subplot(gs[1, 1])
    ax_scatter.scatter(b_est, v_est, alpha=0.7, color='darkorange')
    ax_scatter.axhline(0, color='red', linestyle='--')
    ax_scatter.set_xlabel("Valeurs prédites")
    ax_scatter.set_ylabel("Résidus")
    ax_scatter.set_title("Résidus vs. Valeurs prédites")
    
    # 5. Encadré avec les statistiques – Ligne suivante (s'étend sur 2 colonnes)
    ax_stats = fig.add_subplot(gs[2, :])
    ax_stats.axis('off')  # Masquer les axes
    texte_stats = (
        f"Moyenne    : {moyenne:.4f}\n"
        f"Variance   : {variance:.4f}\n"
        f"Écart-type : {ecart_type:.4f}\n"
        f"Skewness   : {skewness:.4f}\n"
        f"Kurtosis   : {kurtosis:.4f}"
    )
    ax_stats.text(0.5, 0.5, texte_stats, transform=ax_stats.transAxes,
                  fontsize=14, verticalalignment='center', horizontalalignment='center',
                  bbox=dict(facecolor='wheat', edgecolor='black', boxstyle='round,pad=1'))
    ax_stats.set_title("Statistiques des résidus", fontsize=16)
    
    # 6. Encadré avec les informations de position – (optionnel)
    if nrows == 4:
        # Calcul de la distance entre la position estimée et la position du header RINEX
        dist = np.sqrt(np.sum((P_est - P_rnx_header)**2))
        # Calcul des coordonnées ENU locales via la fonction toolCartLocGRS80
        E, N, U = tools.toolCartLocGRS80(P_rnx_header[0], P_rnx_header[1], P_rnx_header[2],
                                          P_est[0], P_est[1], P_est[2])
        texte_extra = (
            f"Distance entre la position estimée et la position initiale du header RINEX : {dist:.4f}\n"
            f"Coordonnées ENU locales :\n"
            f"  Est (E) : {E:.4f}\n"
            f"  Nord (N): {N:.4f}\n"
            f"  Haut (U) : {U:.4f}"
        )
        ax_extra = fig.add_subplot(gs[3, :])
        ax_extra.axis('off')
        ax_extra.text(0.5, 0.5, texte_extra, transform=ax_extra.transAxes,
                      fontsize=14, verticalalignment='center', horizontalalignment='center',
                      bbox=dict(facecolor='lightcyan', edgecolor='black', boxstyle='round,pad=1'))
        ax_extra.set_title("Informations de position", fontsize=16)
    
    # Titre global de la figure si fourni
    if figure_title is not None:
        fig.suptitle(figure_title, fontsize=20)
        plt.subplots_adjust(top=0.92)
    
    # Ajuster la mise en page
    plt.tight_layout(rect=[0, 0, 1, 0.95])
    
    # Sauvegarder la figure si un chemin est fourni
    if save_path is not None:
        fig.savefig(save_path)
        print(f"Figure sauvegardée dans : {save_path}")
    
    plt.show()
    return fig



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 14 23:31:09 2025

@author: psakic
"""
import datetime
import re

import numpy as np
import os
import datetime as dt

import logging

log = logging.getLogger("geodezyx")


def read_antex(filepath, ant_type=None):
    """
    Parse an ANTEX file and extract antenna information.

    Parameters
    ----------
    filepath : str
        Path to the ANTEX file to be read.
    ant_type : str, optional
        Specific antenna type to filter. If provided, only data for this antenna type will be included.

    Returns
    -------
    dict
        A dictionary containing parsed antenna data. The structure includes:
        - HEADER: List of header lines (if `header_in_output` is True).
        - Antenna type as keys, with corresponding metadata and frequency-specific data as values.

    Note
    ----

    The `<Antenna Type>` key in the returned dictionary differs for satellite and non-satellite antennas:

    - **Satellite Antennas**: The key is `SVN`_`VALID FROM`_`VALID UNTIL`.
        Example: `"G042"`

    - **Receiver Antennas**: The key is `TYPE`.
        Example: `"TPSPG_A1+GP     NONE"`

    If the `<Antenna Type>` key already exists, it is incremented with a `_01`...`_nn` suffix


    Structure of the returned dictionary:

    dict  # Root dictionary
    ├── HEADER (list of str)  # List of header lines from the ANTEX file.
    ├── ANTS (dict)  # Dictionary containing antenna data.
    │   ├── <Antenna Type> (str)  # Key representing the antenna type (e.g., "AS-ANT3BCAL").
    │   │   ├── TYPE (str)  # Antenna type (e.g., "AS-ANT3BCAL").
    │   │   ├── SERIAL (str)  # Serial number of the antenna.
    │   │   ├── METH (str)  # Method information (e.g., calibration method).
    │   │   ├── DAZI (float)  # Azimuth increment in degrees.
    │   │   ├── ZEN (list of float)  # Zenith angle range and increment [ZEN1, ZEN2, DZEN].
    │   │   ├── NFREQ (int)  # Number of frequencies for the antenna.
    │   │   ├── SINEX CODE (str, optional)  # SINEX code for the antenna (optional).
    │   │   ├── COMMENT (list of str, optional)  # List of comments related to the antenna (optional).
    │   │   ├── VALID FROM (datetime, optional)  # Start of the validity period (optional).
    │   │   ├── VALID UNTIL (datetime, optional)  # End of the validity period (optional).
    │   │   ├── FREQS (dict)  # Dictionary containing frequency-specific data.
    │   │   │   ├── <Frequency> (str)  # Key representing the frequency (e.g., "G01").
    │   │   │   │   ├── PCO (list of float)  # Phase center offset [NORTH, EAST, UP].
    │   │   │   │   ├── NOAZI (numpy array)  # Elevation-dependent phase center variations.
    │   │   │   │   ├── AZI (numpy array)  # Azimuth-dependent phase center variations.

    """
    atx_dic = {"ANTS": {}, "HEADER": []}
    ant_dic = None
    freq_dic = None
    freq = None
    header = []
    skip_to_next_ant = False

    # Read the file content
    with open(filepath, "r") as f:
        lines = f.readlines()

    in_header = True
    for line in lines:
        label = line[60:].strip()

        # Process header lines
        if in_header:
            atx_dic["HEADER"].append(line)
            if label == "END OF HEADER":
                in_header = False
            continue

        # Start a new antenna block
        if label == "START OF ANTENNA":
            skip_to_next_ant = False
            ant_dic = {
                "COMMENT": [],
                "FREQS": {},
                "METH": "",
                "TYPE": "",
                "SERIAL": "",
                "SVN": "",
                "COSPAR": "",
                "SINEX CODE": "",
                "VALID FROM": None,
                "VALID UNTIL": None,
                "NFREQ": 0,
                "ZEN": 0,
            }

        # Skip processing if the antenna type does not match
        if skip_to_next_ant:
            continue

        # Parse antenna type and serial number
        elif label == "TYPE / SERIAL NO":
            if ant_type and not line[:20].strip().startswith(ant_type):
                skip_to_next_ant = True

            ant_dic["TYPE"] = line[0:20].strip()
            ant_dic["SERIAL"] = line[20:40].strip()
            ant_dic["SVN"] = line[40:50].strip()
            ant_dic["COSPAR"] = line[50:60].strip()

        elif label == "TYPE / SERIAL NO":
            if ant_type and not line[:20].strip().startswith(ant_type):
                skip_to_next_ant = True

            ant_dic["TYPE"] = line[0:20].strip()
            ant_dic["SERIAL"] = line[20:40].strip()
            ant_dic["SVN"] = line[40:50].strip()
            ant_dic["COSPAR"] = line[50:60].strip()

        # Parse azimuth increment
        elif label == "DAZI":
            ant_dic["DAZI"] = float(line[:6])

        # Parse zenith angle range and increment
        elif label == "ZEN1 / ZEN2 / DZEN":
            ant_dic["ZEN"] = list(map(float, line[:60].split()))

        # Parse comments
        elif label == "COMMENT":
            ant_dic["COMMENT"].append(line[:60])

        # Parse the number of frequencies
        elif label == "# OF FREQUENCIES":
            ant_dic["NFREQ"] = int(line[:20].strip())

        # Parse Validity dates
        elif label == "VALID FROM":
            ant_dic["VALID FROM"] = dt.datetime(
                *list(map(int, line[:35].strip().split()))
            ) + dt.timedelta(microseconds=0)

        elif label == "VALID UNTIL":
            ant_dic["VALID UNTIL"] = dt.datetime(
                *list(map(int, line[:35].strip().split()))
            ) + dt.timedelta(microseconds=999999)

        # Parse SINEX code
        elif label == "SINEX CODE":
            ant_dic["SINEX CODE"] = line[:20].strip()

        # Parse method, author, and date
        elif label == "METH / BY / # / DATE":
            ant_dic["METH"] = line[0:20].strip()
            ant_dic["BY"] = line[20:40].strip()
            ant_dic["NCAL"] = line[40:46].strip()
            ant_dic["DATE"] = line[50:60].strip()

        # Start a new frequency block
        elif label == "START OF FREQUENCY":
            freq = line[:20].strip()
            ant_dic["FREQS"][freq] = {"AZI": [], "NOAZI": "", "PCO": []}
            freq_dic = ant_dic["FREQS"][freq]

        # End the current frequency block
        elif label == "END OF FREQUENCY":
            freq = None

        # Parse phase center offset (PCO) values
        elif freq and label == "NORTH / EAST / UP":
            freq_dic["PCO"] = list(map(float, line[:60].split()))

        # Parse elevation-dependent phase center variations (PCV) for NOAZI
        elif freq and line.strip().startswith("NOAZI"):
            values = np.array(list(map(float, line[:].split()[1:])))
            freq_dic["NOAZI"] = values

        # Parse azimuth-dependent PCV values
        elif freq and not line.strip().startswith("NOAZI"):
            values = np.array(list(map(float, line[:].split())))
            freq_dic["AZI"].append(values)

        # End the current antenna block
        elif label == "END OF ANTENNA":
            for f in ant_dic["FREQS"].keys():
                if ant_dic["FREQS"][f]["AZI"]:
                    ant_dic["FREQS"][f]["AZI"] = np.vstack(ant_dic["FREQS"][f]["AZI"])

            ### must define a specific antenna key kant
            if ant_dic["COSPAR"]:  # Antenna is a satellite one
                s = (
                    ant_dic["VALID FROM"].strftime("%y%m%d")
                    if ant_dic["VALID UNTIL"]
                    else "000000"
                )
                e = (
                    ant_dic["VALID UNTIL"].strftime("%y%m%d")
                    if ant_dic["VALID UNTIL"]
                    else "999999"
                )
                kant = "_".join((ant_dic["SVN"], s, e))
            else:  # Antenna is a receiver one
                kant = ant_dic["TYPE"]

            ### we must increment antenna key if already one defined
            ## (not very useful and not really tested)
            if kant in ant_dic.keys():
                if re.search(".*_[0-9]{2}$", kant):
                    kant = kant[:-3] + "_" + str(int(kant[:-2]) + 1)
                else:
                    kant = kant + "_01"

            atx_dic["ANTS"][kant] = ant_dic
            ant_dic = None

    return atx_dic


def l_atx_std(val_inp, label_inp, newline=True):
    """
    Format a standard ANTEX line with a value and label.

    Parameters
    ----------
    val_inp : str
        The value to be written in the line.
    label_inp : str
        The label to be written in the line.

    Returns
    -------
    str
        Formatted ANTEX line.
    """
    nl = "\n" if newline else ""
    return f"{val_inp:60}{label_inp:20}" + nl


def l_atx_cal(val_inp, noazi=True, newline=True):
    """
    Format calibration data for ANTEX.

    Parameters
    ----------
    val_inp : list
        Calibration values to be written.
    noazi : bool, optional
        If True, format as NOAZI calibration. Default is True.

    Returns
    -------
    str
        Formatted calibration data.
    """
    nl = "\n" if newline else ""
    if noazi:
        n = len(val_inp)
        lin = "   NOAZI" + n * "{:8.2f}" + nl
        return lin.format(*val_inp)
    else:
        lin_stk = []
        for lmat in val_inp:
            n = len(lmat)
            lin = "{:8.1f}" + (n - 1) * "{:8.2f}"
            lin_stk.append(lin.format(*lmat))
        return "\n".join(lin_stk) + nl


def write_antex(atx_dic_inp, dir_out, fname_out="out.atx", erase_header=False):
    """
    Write antenna calibration information to an ANTEX file.

    Parameters
    ----------
    atx_dic_inp : dict
        Dictionary containing antenna calibration
        data to be written to the ANTEX file.
        See `read_antex` for dictionnary structure
    dir_out : str
        Directory where the output ANTEX file will be saved.
    fname_out : str, optional
        Name of the output ANTEX file. Default is 'out.atx'.
    erase_header : bool, optional
        If True, the header in the input dictionary will be replaced with a default header. Default is False.

    Notes
    -----
    - The function formats the antenna data into the ANTEX file format and writes it to the specified directory.
    - If `erase_header` is True, a default header is used instead of the one provided in the input dictionary.

    Returns
    -------
    str
        Path to the generated ANTEX file.
    """

    lout = []

    # Add header if present in the input dictionary
    if "HEADER" in atx_dic_inp and not erase_header:
        lhead = atx_dic_inp["HEADER"]
    else:
        lhead = [
            l_atx_std("     1.4            M", "ANTEX VERSION / SYST"),
            l_atx_std("A", "PCV TYPE / REFANT"),
            l_atx_std("", "END OF HEADER"),
        ]

    gyx_hd = l_atx_std(
        "HANDELED WITH GEODEZYX TOOLBOX - github.com/IPGP/geodezyx", "COMMENT"
    )

    if not "END OF HEADER" in lhead[-1]:
        lhead.append(60 * " " + "END OF HEADER       \n")

    lhead = lhead[:-1] + [gyx_hd] + lhead[-1:]
    lout.append("".join(lhead))

    # Iterate over antennas in the input dictionary
    for kant, ant in atx_dic_inp["ANTS"].items():

        # Write antenna start label
        label = "START OF ANTENNA"
        lout.append(l_atx_std("", label))

        # Write antenna type and serial number
        label = "TYPE / SERIAL NO"
        val = f"{ant['TYPE']:20}{ant['SERIAL']:20}{ant['SVN']:10}{ant['COSPAR']:10}"
        lout.append(l_atx_std(val, label))

        # Write method, date, and other metadata
        label = "METH / BY / # / DATE"
        val = f"{ant['METH']:20}{ant['BY']:20}{ant['NCAL']:>6}    {ant['DATE']:10}"
        lout.append(l_atx_std(val, label))

        label = "DAZI"
        val = f"  {ant['DAZI']:6.1f}"
        lout.append(l_atx_std(val, label))

        label = "ZEN1 / ZEN2 / DZEN"
        val = "  {:6.1f}{:6.1f}{:6.1f}".format(*ant["ZEN"])
        lout.append(l_atx_std(val, label))

        label = "# OF FREQUENCIES"
        nfreq = len(ant["FREQS"])
        if nfreq != ant["NFREQ"]:
            log.info(
                f"# OF FREQUENCIES for {ant['TYPE']} will be updated! old:{ant['NFREQ']} new:{nfreq}"
            )
            ant["NFREQ"] = nfreq
        val = f"{ant['NFREQ']:6}"
        lout.append(l_atx_std(val, label))

        dat_fmt = 5 * "{:6d}" + "{:13.7f}"

        for label in ["VALID FROM", "VALID UNTIL"]:
            if ant[label]:
                d = ant[label]
                datlis = list(d.timetuple())[:6]
                lastmusec = 9 * 10**-7 if d.microsecond == 999999 else 0
                datlis[-1] = datlis[-1] + d.microsecond * 10**-6 + lastmusec
                val = dat_fmt.format(*datlis)
                lout.append(l_atx_std(val, label))

        # Facultative labels
        if "SINEX CODE" in ant.keys():
            label = "SINEX CODE"
            val = f"{ant['SINEX CODE']:60}"
            lout.append(l_atx_std(val, label))

        # Write comments
        if "COMMENT" in ant.keys():
            for com in ant["COMMENT"]:
                label = "COMMENT"
                val = f"{com:60}"
                lout.append(l_atx_std(val, label))

        # Write frequency-specific data
        for kfreq, freq in ant["FREQS"].items():

            label = "START OF FREQUENCY"
            val = f"   {kfreq:57}"
            lout.append(l_atx_std(val, label))

            label = "NORTH / EAST / UP"
            val = "{:10.2f}{:10.2f}{:10.2f}".format(*freq["PCO"])
            lout.append(l_atx_std(val, label))

            if len(freq["NOAZI"]) > 0:
                lout.append(l_atx_cal(freq["NOAZI"], noazi=True))
            if len(freq["AZI"]) > 0:
                lout.append(l_atx_cal(freq["AZI"], noazi=False))

            label = "END OF FREQUENCY"
            val = f"   {kfreq:57}"
            lout.append(l_atx_std(val, label))

        # Write antenna end label
        label = "END OF ANTENNA"
        lout.append(l_atx_std("", label))

    # Write the output to the specified file
    path_out = os.path.join(dir_out, fname_out)
    f_out = open(path_out, "w+")
    for l in lout:
        f_out.write(l)

    return path_out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 23/04/2024 21:31:45

@author: psakic
"""
import datetime as dt
import glob
import itertools
import logging
import os
import pathlib
import re

import pandas as pd

import geodezyx.operational.download_utils as dlutils
from geodezyx import conv

log = logging.getLogger("geodezyx")


def _rnx_obs_rgx(stat, date):
    """
    Generate RINEX observation file regex patterns for both RINEX2 and RINEX3 formats.

    This internal function creates regex patterns to match RINEX observation files
    for a given station and date, supporting both the legacy 8.3 filename convention
    (RINEX2) and the long filename convention (RINEX3).

    Parameters
    ----------
    stat : str
        4-character GNSS station name (e.g., 'ZIMM', 'TLSE').
        Will be converted to lowercase for RINEX2 pattern.
    date : datetime.datetime
        Date for which to generate the regex patterns.

    Returns
    -------
    tuple of (str, str)
        A tuple containing:
        - rnx2rgx : str
            Regex pattern for RINEX2 observation files (8.3 format)
        - rnx3rgx : str
            Regex pattern for RINEX3 observation files (long format)

    Notes
    -----
    The RINEX2 pattern uses the station name in lowercase with a wildcard
    for the file type extension. The RINEX3 pattern follows the standard
    long filename convention with:
    - Country code: wildcard ("...")
    - Data source: wildcard (".")
    - File period: 01D (daily)
    - Data frequency: wildcard ("...")
    - Data type: .O (observation)
    - Format/compression: wildcard (".*")

    Examples
    --------
    ```
    >>> import datetime as dt
    >>> rnx2_pattern, rnx3_pattern = _rnx_obs_rgx('ZIMM', dt.datetime(2020, 1, 1))
    >>> # rnx2_pattern might be: 'zimm001a.20o.*'
    >>> # rnx3_pattern might be: 'ZIMM...._R_20200010000_01D_....O.*'
    ```
    """
    rnx2rgx = conv.statname_dt2rinexname(stat.lower(), date, rnxtype=".*")
    rnx3rgx = conv.statname_dt2rinexname_long(
        stat,
        date,
        country="...",
        data_source=".",
        file_period="01D",
        data_freq="...",
        data_type=".O",
        format_compression=".*",
    )
    return rnx2rgx, rnx3rgx


def _rnx_nav_rgx(stat, date, sys=".", data_source="."):
    """
    Generate RINEX navigation file regex patterns for both RINEX2 and RINEX3 formats.

    This internal function creates regex patterns to match RINEX navigation files
    for a given station and date, supporting both the legacy 8.3 filename convention
    (RINEX2) and the long filename convention (RINEX3). Navigation files contain
    satellite ephemeris and clock correction data.

    Parameters
    ----------
    stat : str
        4-character GNSS station name (e.g., 'BRDC', 'ZIMM').
        For broadcast navigation files, typically use 'BRDC'.
    date : datetime.datetime
        Date for which to generate the regex patterns.
    sys : str, optional
        GNSS system identifier for RINEX3 navigation files. Default is "." (wildcard).
        Common values:
        - "G" : GPS navigation data
        - "R" : GLONASS navigation data
        - "E" : Galileo navigation data
        - "C" : BeiDou navigation data
        - "M" : Mixed/Multi-GNSS navigation data
        - "." : Wildcard for any system
    data_source : str, optional
        Data source identifier for RINEX3 long filenames. Default is "." (wildcard).
        Common values:
        - "R" : Real-time data
        - "S" : Survey/static data
        - "U" : Ultra-rapid data
        - "." : Wildcard for any source

    Returns
    -------
    tuple of (str, str)
        A tuple containing:
        - rnx2rgx : str
            Regex pattern for RINEX2 navigation files (8.3 format)
        - rnx3rgx : str
            Regex pattern for RINEX3 navigation files (long format)

    Notes
    -----
    The RINEX2 pattern uses the station name in lowercase with a wildcard
    for the file type extension. The RINEX3 pattern follows the standard
    long filename convention for navigation files with:
    - Country code: wildcard ("...")
    - Data source: configurable (default wildcard)
    - File period: 01D (daily)
    - Data frequency: empty ("") for navigation files
    - Data type: sys + "N" (Navigation)
    - Format/compression: wildcard (".*")

    Examples
    --------
    >>> import datetime as dt
    >>> rnx2_pattern, rnx3_pattern = _rnx_nav_rgx('BRDC', dt.datetime(2020, 1, 1))
    >>> # rnx2_pattern might be: 'brdc001a.20n.*'
    >>> # rnx3_pattern might be: 'BRDC...._._20200010000_01D_.N.*'

    >>> # For GPS-specific navigation files
    >>> rnx2_gps, rnx3_gps = _rnx_nav_rgx('BRDC', dt.datetime(2020, 1, 1), sys="G", data_source="R")
    >>> # rnx3_gps might be: 'BRDC...._R_20200010000_01D_GN.*'
    """
    rnx2rgx = conv.statname_dt2rinexname(stat.lower(), date, rnxtype=".*")
    rnx3rgx = conv.statname_dt2rinexname_long(
        stat,
        date,
        country="...",
        data_source=data_source,
        file_period="01D",
        data_freq="",
        data_type=sys + "N",
        format_compression=".*",
    )

    return rnx2rgx, rnx3rgx


def _generic_server(stat, date, urlserver, urlsuffix=None):
    """
    Generate RINEX file URLs for a generic FTP server structure.

    This internal function creates download URLs for both RINEX2 and RINEX3 observation
    files following a standard FTP server directory structure: server/year/doy/filename.
    This is the most common layout used by many GNSS data centers.

    Parameters
    ----------
    stat : str
        4-character GNSS station name (e.g., 'ZIMM', 'TLSE').
        Station name case will be handled appropriately for each RINEX format.
    date : datetime.datetime
        Date for which to generate the URLs.
        Used to construct the year/doy directory path and filename patterns.
    urlserver : str
        Base FTP server URL (e.g., 'ftp://example.com/data/').
        Should include the protocol and base path to the RINEX data directory.

    Returns
    -------
    dict
        Dictionary mapping RINEX version numbers to their respective URLs:
        - urldic[2] : str - URL for RINEX2 observation file
        - urldic[3] : str - URL for RINEX3 observation file

    Notes
    -----
    The function follows the standard GNSS data center directory structure:
    - urlserver/YYYY/DOY/filename
    Where:
    - YYYY is the 4-digit year
    - DOY is the 3-digit day of year (001-366)
    - filename follows RINEX2 or RINEX3 naming conventions

    This generic structure is used by servers like IGS SOPAC, IGN, and others.
    Some servers (like CDDIS) have custom structures and require specialized functions.

    Examples
    --------
    >>> import datetime as dt
    >>> urls = _generic_server('ZIMM', dt.datetime(2020, 1, 15), 'ftp://example.com/data/')
    >>> # urls[2] might be: 'ftp://example.com/data/2020/015/zimm015a.20o.*'
    >>> # urls[3] might be: 'ftp://example.com/data/2020/015/ZIMM...._R_20200150000_01D_....O.*'
    """
    rnx2rgx, rnx3rgx = _rnx_obs_rgx(stat, date)

    if not urlsuffix:
        urlsuffix = ""

    ### generate urls
    urldir = str(os.path.join(urlserver, str(date.year), conv.dt2doy(date), urlsuffix))
    rnx2url = os.path.join(urldir, rnx2rgx)
    rnx3url = os.path.join(urldir, rnx3rgx)

    ### generate output urldic, key 2 and 3 are for rinex version
    urldic = dict()
    urldic[2] = rnx2url
    urldic[3] = rnx3url

    return urldic

def igs_sopac_server(stat, date):
    # plante si trop de requete
    urlserver = "ftp://garner.ucsd.edu/pub/rinex/"
    urldic = _generic_server(stat, date, urlserver)
    return urldic


def igs_cddis_server(stat, date):
    # plante si trop de requete
    urlserver = "ftp://gdc.cddis.eosdis.nasa.gov/gps/data/daily/"

    # we can not use _generic_server here because of the specific server path structure

    ### generate regex
    rnx2rgx, rnx3rgx = _rnx_obs_rgx(stat, date)

    ### generate urls
    urldir = os.path.join(
        urlserver, str(date.year), conv.dt2doy(date), date.strftime("%y") + "d"
    )
    rnx2url = os.path.join(urldir, rnx2rgx)
    rnx3url = os.path.join(urldir, rnx3rgx)

    ### generate output urldic, key 2 and 3 are for rinex version
    urldic = dict()
    urldic[2] = rnx2url
    urldic[3] = rnx3url

    return urldic


def igs_ign_server(stat, date):
    # plante si trop de requete
    urlserver = "ftp://igs.ign.fr/pub/igs/data/"
    urldic = _generic_server(stat, date, urlserver)
    return urldic


def igs_ign_ensg_server(stat, date):
    # plante si trop de requete
    urlserver = "ftp://igs.ensg.eu/pub/igs/data/"
    urldic = _generic_server(stat, date, urlserver)
    return urldic


def igs_bkg_server(stat, date):
    urlserver = "ftp://igs-ftp.bkg.bund.de/IGS/obs/"
    # ftp://igs-ftp.bkg.bund.de/IGS/obs/2024/082/IGS00WRD_R_20240820000_01D_MN.rnx.gz

    urldic = _generic_server(stat, date, urlserver)
    return urldic


def nav_rob_server(stat, date):
    urlserver = "ftp://epncb.oma.be/pub/obs/BRDC/"

    # can not use _generic_server here because of the specific server path structure / file name

    ### generate regex
    rnx2rgx, rnx3rgx = _rnx_nav_rgx(stat, date)  ### NAV RNX HERE !!!

    ### generate urls
    urldir = os.path.join(urlserver, str(date.year))  ## NO DOY FOR THIS ONE !!!
    rnx2url = os.path.join(urldir, rnx2rgx)
    rnx3url = os.path.join(urldir, rnx3rgx)

    ### generate output urldic, key 2 and 3 are for rinex version
    urldic = {}
    urldic[2] = rnx2url
    urldic[3] = rnx3url

    return urldic


def sonel_server(stat, date):
    urlserver = "ftp://ftp.sonel.org/gps/data/"
    urldic = _generic_server(stat, date, urlserver)
    return urldic


def rgp_server(stat, date):
    urlserver = "ftp://rgpdata.ign.fr/pub/data/"
    urldic = _generic_server(stat, date, urlserver, "data_30")
    return urldic


def rgp_ensg_server(stat, date):
    urlserver = "ftp://rgpdata.ensg.eu/pub/data/"
    urldic = _generic_server(stat, date, urlserver, "data_30")
    return urldic

def spotgins_eost_server(stat, date):
    urlserver = "http://loading.u-strasbg.fr/SPOTGINS/TEST/rinex/"
    urldic = _generic_server(stat, date, urlserver)
    return urldic

def euref_server(stat, date):
    urlserver = "ftp://epncb.oma.be/pub/obs/"

    # can not use _generic_server here because of upper case RINEX 2 names
    rnx2rgx, rnx3rgx = _rnx_obs_rgx(stat, date)
    rnx2rgx = rnx2rgx.upper()

    ### generate urls
    urldir = str(os.path.join(urlserver, str(date.year), conv.dt2doy(date)))
    rnx2url = os.path.join(urldir, rnx2rgx)
    rnx3url = os.path.join(urldir, rnx3rgx)

    ### generate output urldic, key 2 and 3 are for rinex version
    urldic = dict()
    urldic[2] = rnx2url
    urldic[3] = rnx3url

    return urldic


def nav_bkg_server(stat, date):
    urlserver = "ftp://igs-ftp.bkg.bund.de/IGS/BRDC/"
    # ftp://igs-ftp.bkg.bund.de/IGS/BRDC/2024/082/BRDC00WRD_S_20240820000_01D_MN.rnx.gz

    ### generate regex
    rnx2rgx, rnx3rgx = _rnx_nav_rgx(
        stat, date, sys="M", data_source="S"
    )  ### NAV RNX HERE !!!

    ### generate urls
    urldir = os.path.join(urlserver, str(date.year), conv.dt2doy(date))
    rnx3url = os.path.join(urldir, rnx3rgx)

    urldic = dict()
    urldic[3] = rnx3url

    return urldic


############ not adapted yet after april 2024 mods
def igs_cddis_nav_server_legacy(stat, date):
    # table_proto privilegier
    urlserver = "ftp://cddis.gsfc.nasa.gov/gps/data/daily/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date, "n.Z")
    url = os.path.join(
        urlserver, str(date.year), conv.dt2doy(date), date.strftime("%y") + "n", rnxname
    )
    return url


def rgp_ign_smn_1_hz_server_legacy(stat, date):
    urlserver = "ftp://rgpdata.ign.fr/pub/data/"

    urls = []

    for h in range(24):
        date_session = date
        date_session = date_session.replace(hour=h)

        log.info("%s session %s", date_session, h)
        rnxname = conv.statname_dt2rinexname(
            stat.lower(), date_session, session_a_instead_of_daily_session=True
        )
        url = os.path.join(
            urlserver, str(date.year), conv.dt2doy(date), "data_1", rnxname
        )

        urls.append(url)

    return urls


def unavco_server_legacy(stat, date):
    urlserver = "ftp://data-out.unavco.org/pub/rinex"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, "obs", str(date.year), conv.dt2doy(date), rnxname)
    return url


def renag_server_legacy(stat, date):
    urlserver = "ftp://renag.unice.fr/data/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), rnxname)
    return url


def uwiseismic_server_legacy(stat, date, user="", passwd=""):
    urlserver = "ftp://www2.uwiseismic.com/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, "rinex", str(date.year), conv.dt2doy(date), rnxname)
    return url, user, passwd


def orpheon_server_legacy(stat, date, user="", passwd=""):
    urlserver = "ftp://renag.unice.fr/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), rnxname)
    return url, user, passwd


def ovsg_server_legacy(stat, date, user="", passwd=""):
    if dt.datetime(2009, 1, 1) <= date <= dt.datetime(2014, 2, 10):
        urlserver = (
            "http://webobs.ovsg.univ-ag.fr/rawdata/GPS-GPSDATA.backtemp_20140210/"
        )
    else:
        urlserver = "http://webobs.ovsg.univ-ag.fr/rawdata/GPS/GPSDATA/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), "rinex", rnxname)
    return url, user, passwd


def geoaus_server_legacy(stat, date):
    """Geosciences Australia
    ex : ftp://ftp.ga.gov.au/geodesy-outgoing/gnss/data/daily/2010/10063/"""
    urlserver = "ftp://ftp.ga.gov.au/geodesy-outgoing/gnss/data/daily/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(
        urlserver, str(date.year), date.strftime("%y") + conv.dt2doy(date), rnxname
    )
    return url


def ens_fr_legacy(stat, date):
    urlserver = "ftp://gnss.ens.fr/pub/public/crl/GPS/rinex/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), rnxname)
    return url


def _server_select(datacenter, site, curdate):
    mode1hz = False
    secure_ftp = False
    urldic = dict()
    if datacenter in ("igs_cddis", "igs"):
        urldic = igs_cddis_server(site, curdate)
        secure_ftp = True
    elif datacenter == "igs_sopac":
        urldic = igs_sopac_server(site, curdate)
    elif datacenter == "igs_ign":
        urldic = igs_ign_server(site, curdate)
    elif datacenter == "igs_ign_ensg":
        urldic = igs_ign_ensg_server(site, curdate)
    elif datacenter == "sonel":
        urldic = sonel_server(site, curdate)
    elif datacenter == "euref":
        urldic = euref_server(site, curdate)
    elif datacenter == "igs_bkg":
        urldic = igs_bkg_server(site, curdate)
    elif datacenter in ("nav", "brdc"):
        urldic = nav_rob_server(site, curdate)
    elif datacenter in ("nav_rt", "brdc_rt"):
        urldic = nav_bkg_server(site, curdate)
    elif datacenter == "rgp":
        urldic = rgp_server(site, curdate)
    elif datacenter == "rgp_ensg":
        urldic = rgp_ensg_server(site, curdate)
    elif datacenter == "spotgins_eost":
        urldic = spotgins_eost_server(site, curdate)
    # elif datacenter == 'rgp':
    #     urldic = rgp_ign_smn_server_legacy(site, curdate)
    # elif datacenter == 'rgp_mlv':
    #     urldic = rgp_ign_mlv_server(site, curdate)
    # elif datacenter == 'rgp_1Hz':
    #     urls = rgp_ign_smn_1_hz_server_legacy(site, curdate)
    #     mode1hz = True
    # elif datacenter == 'renag':
    #     urldic = renag_server(site, curdate)
    # elif datacenter == 'orpheon':
    #     urldic = orpheon_server_legacy(site, curdate)
    # elif datacenter == 'uwiseismic':
    #     urldic = uwiseismic_server(site, curdate)
    # elif datacenter == 'ovsg':
    #     urldic = ovsg_server_legacy(site, curdate)
    # elif datacenter == 'unavco':
    #     urldic = unavco_server_legacy(site, curdate)
    # elif datacenter == 'geoaus':
    #     urldic = geoaus_server_legacy(site, curdate)
    # elif datacenter == 'ens_fr':
    #     urldic = ens_fr_legacy(site, curdate)
    else:
        log.warning("unkwn server dic in the dico, skip ...")
        return None, None, None

    return urldic, secure_ftp, mode1hz

def effective_save_dir(parent_archive_dir, stat, date, archtype="stat"):
    """
    INTERNAL_FUNCTION

    archtype =
        stat
        stat/year
        stat/year/doy
        year/doy
        year/stat
        week/dow
        OR only '/' for a dirty saving in the parent folder
        ... etc ..."""
    if archtype == "/":
        return parent_archive_dir

    if len(archtype) > 0 and archtype.startswith("/"):
        log.warning("The archive type starts with /, remove it to avoid error")

    out_save_dir = parent_archive_dir
    fff = archtype.split("/")
    year = str(date.year)
    doy = conv.dt2doy(date)
    _, _ = year, doy  ## simply to remove the unused linter warning...
    week, dow = conv.dt2gpstime(date)
    for f in fff:
        out_save_dir = os.path.join(out_save_dir, eval(f))
    return out_save_dir


def rnx_regex_indir(rnx_regex, dir_files_list):
    """
    Match files in a directory against a given regex pattern.

    Parameters
    ----------
    rnx_regex : str
        Regex pattern to match filenames.
    dir_files_list : list of str
        List of filenames in the directory.

    Returns
    -------
    str or None
        The first matching filename, or None if no match is found.
    """
    matches = [file for file in dir_files_list if re.search(rnx_regex, file)]
    return matches[0] if matches else None


def crawl_ftp_files(
    table,
    sftp="auto",
    user=None,
    passwd=None,
    path_ftp_crawled_files_save=None,
    path_all_ftp_files_save=None,
    force=False,
):
    """
    Crawl FTP servers to find available RINEX files and update download table.

    This function performs an optimized FTP crawl by reusing connections and
    minimizing directory changes. It checks for existing local files, connects
    to FTP servers, lists remote files, and updates the table with availability
    status and actual file URLs.

    Parameters
    ----------
    table : pd.DataFrame
        Input table containing RINEX download metadata with columns:
        - 'host': FTP server hostname
        - 'dir': Remote directory path
        - 'outdir': Local output directory
        - 'rnxrgx': RINEX filename regex pattern
        - 'sftp': Boolean indicating if SFTP should be used
        - 'crawled': Boolean indicating if already crawled
    sftp : str or bool, optional
        SFTP mode setting. Default is 'auto'.
        - 'auto': Use the 'sftp' column value from each table row
        - True/False: Force SFTP on/off for all connections
    user : str, optional
        FTP username. Default is None (anonymous).
    passwd : str, optional
        FTP password. Default is None (anonymous).
    path_ftp_crawled_files_save : str, optional
        Path to save the crawled files table as CSV.
        If None, no file is saved.
    path_all_ftp_files_save : str, optional
        Path to save all discovered FTP files as CSV.
        If None, no file is saved.
    force : bool, optional
        Force re-download even if files exist locally. Default is False.

    Returns
    -------
    tuple of (pd.DataFrame, pd.Series, pd.Series)
        - table_use : pd.DataFrame
            Updated table with crawl results, including new columns:
            - 'ok_dwl': Boolean indicating file is available for download
            - 'ok_loc': Boolean indicating file exists locally
            - 'rnxnam': Actual filename found on server
            - 'url_true': Complete FTP URL for download
        - all_ftp_files : pd.Series
            All files discovered on FTP servers with full URLs
        - all_loc_files : pd.Series
            Local file paths for files that already exist

    Notes
    -----
    The function implements several optimizations:
    - Reuses FTP connections when possible (same host)
    - Reconnects every 50 operations to avoid timeouts
    - Caches local and remote directory listings
    - Only changes directories when necessary
    - Saves intermediate results for recovery

    The crawling process:
    1. Checks for existing local files first
    2. Connects to FTP server when host changes
    3. Lists remote directory contents when directory changes
    4. Matches files using regex patterns
    5. Updates table with availability status
    6. Generates download URLs for available files

    Examples
    --------
    >>> import pandas as pd
    >>> table = pd.DataFrame({
    ...     'host': ['ftp.example.com'],
    ...     'dir': ['/data/2020/001'],
    ...     'outdir': ['/local/data'],
    ...     'rnxrgx': ['station001a.20o.*'],
    ...     'sftp': [False],
    ...     'crawled': [False]
    ... })
    >>> crawled_table, all_files, local_files = crawl_ftp_files(table)
    """

    def _save_crawled_files(table_inp):
        """Save crawled files table to CSV if path is provided."""
        if path_ftp_crawled_files_save:
            table_inp.to_csv(path_ftp_crawled_files_save)

    def _get_and_save_all_ftp_files(all_ftp_files_stk_inp):
        """Concatenate and save all discovered FTP files."""
        if all_ftp_files_stk_inp:
            all_ftp_files_out = pd.concat(all_ftp_files_stk_inp)
            all_ftp_files_out.reset_index(drop=True, inplace=True)
        else:
            all_ftp_files_out = pd.Series([], dtype=str)

        if path_all_ftp_files_save:
            all_ftp_files_out.to_csv(path_all_ftp_files_save)

        return all_ftp_files_out

    table_use = table.copy()

    # Initialize loop variables
    prev_host = ""  # Track previous host to reuse connections
    prev_dir = ""  # Track previous directory to avoid unnecessary changes
    ftp_files_list = []  # Cache of current directory file listing
    all_ftp_files_stk = []  # Stack to collect all discovered files
    local_files_lis = []  # Cache of current local directory listing
    count_loop = 0  # Counter for connection refresh
    count_nmax = 50  # Maximum operations before reconnecting
    ftpobj = None  # Current FTP connection object

    for irow, row in table_use.iterrows():
        # Skip rows already crawled
        if row["crawled"]:
            continue

        # Check local files when directory changes
        if prev_dir != row["outdir"] or irow == 0:
            local_files_lis = glob.glob(row["outdir"] + "/*")

        # Check if file already exists locally
        rnxlocal = rnx_regex_indir(row["rnxrgx"], local_files_lis)

        # Skip if file exists locally and not forcing download
        if rnxlocal and os.path.getsize(rnxlocal) > 0:
            rnxloc_bn = os.path.basename(rnxlocal)
            if not force:
                log.info("%s already exists locally ;)", rnxloc_bn)
                table_use.loc[irow, "ok_loc"] = True
                table_use.loc[irow, "ok_dwl"] = False
                table_use.loc[irow, "rnxnam"] = os.path.basename(rnxlocal)
                continue
            else:  # force mode
                log.info(
                    "%s already exists locally, but re-download forced",
                    rnxloc_bn,
                )
                table_use.loc[irow, "ok_loc"] = False
                table_use.loc[irow, "ok_dwl"] = True
                table_use.loc[irow, "rnxnam"] = os.path.basename(rnxlocal)

        count_loop += 1

        # Create new FTP connection if needed
        if row["host"] != prev_host or count_loop > count_nmax or count_loop == 1:
            if ftpobj:
                ftpobj.close()

            # Determine SFTP mode: use row value if 'auto', otherwise use parameter
            sftp_use = bool(row["sftp"]) if sftp == "auto" else sftp
            ftpobj, _ = dlutils.ftp_objt_create(
                secure_ftp_inp=sftp_use,
                host=row["host"],
                user=user,
                passwd=passwd,
            )
            prev_host = row["host"]

            # Save intermediate results and reset counter on reconnection
            if count_loop > count_nmax:
                count_loop = 0
                _save_crawled_files(table_use)
                _get_and_save_all_ftp_files(all_ftp_files_stk)

        # Get file list when directory changes
        if prev_dir != row["dir"] or irow == 0:
            log.info("chdir " + row["dir"])
            ftpobj.cwd("/")  # Reset to root directory first

            try:
                ftpobj.cwd(row["dir"])
                ftp_files_list = dlutils.ftp_dir_list_files(ftpobj)

                # Save all FTP files for reporting with full URLs
                ftp_files_series = pd.Series(ftp_files_list)
                ftp_files_series = (
                    os.path.join("ftp://", row["host"], row["dir"])
                    + "/"
                    + ftp_files_series
                )
                all_ftp_files_stk.append(ftp_files_series)

            except Exception as e:
                log.warning("unable to chdir to %s, exception %s", row["dir"], e)
                ftp_files_list = []

            prev_dir = row["dir"]

        # Check if file exists on server using regex pattern
        rnx_return = rnx_regex_indir(row["rnxrgx"], ftp_files_list)

        # Update table based on file availability
        if rnx_return:
            table_use.loc[irow, "rnxnam"] = rnx_return
            table_use.loc[irow, "ok_dwl"] = True
            log.info(rnx_return + " found on server :)")
        else:
            table_use.loc[irow, "rnxnam"] = ""
            table_use.loc[irow, "ok_dwl"] = False
            log.warning(row["rnxrgx"] + " not found on server :(")

        table_use.loc[irow, "crawled"] = True

    # Generate URLs for downloadable files (available remotely but not locally)
    rnx_ok_dwl = table_use["ok_dwl"] & ~table_use["ok_loc"]
    table_use.loc[rnx_ok_dwl, "url_true"] = table_use.loc[rnx_ok_dwl].apply(
        lambda x: os.path.join("ftp://", x["host"], x["dir"], x["rnxnam"]), axis=1
    )

    # Save final results
    _save_crawled_files(table_use)
    all_ftp_files = _get_and_save_all_ftp_files(all_ftp_files_stk)

    # Clean up FTP connection
    if ftpobj:
        ftpobj.close()

    # Generate local file paths for files that exist locally
    rnx_ok_loc = table_use["ok_loc"]
    if rnx_ok_loc.sum() > 0:
        all_loc_files = pd.Series(
            table_use.loc[rnx_ok_loc].apply(
                lambda e: os.path.join(e["outdir"], e["rnxnam"]), axis=1
            )
        )
    else:
        all_loc_files = pd.Series([])

    return table_use, all_ftp_files, all_loc_files


def download_gnss_rinex(
    statdico,
    output_dir,
    startdate,
    enddate,
    archtype='year/doy',
    parallel_download=4,
    user="anonymous",
    passwd="anonymous@isp.com",
    path_ftp_crawled_files_save=None,
    path_ftp_crawled_files_load=None,
    skip_crawl=False,
    path_all_ftp_files_save=None,
    quiet_mode=False,
    final_archive_for_sup_check=None,
    force=False,
    no_rnx2=False,
    no_rnx3=False,
):
    """
    Parameters
    ----------
    statdico : dict
        a statdico is a dictionary associating Archives Centers to list of stations

        Exemple:
            >>> statdico['archive center 1'] = ['STA1','STA200XXX','STA3', ...]
            >>> statdico['archive center 2'] = ['STA200XXX','STA1','STA4', ...]

        the supported archive center are (april 2024):
            * igs_cddis or igs (CDDIS data center)
            * igs_sopac (for the SOPAC/UCSD/SIO data center, but not very reliable)
            * igs_ign (IGN's data center, main server at St Mandé)
            * igs_ign_ensg (IGN's data center, secondary server at ENSG, Marne-la-Vallée)
            * igs_bkg (BKG's IGS data center, for the IGS stations)
            * sonel
            * euref (EPN data center hosted at ROB)
            * nav or brdc as archive center allows to download nav files (using 'BRDC' as station name)
            * from the ROB server, using GOP files
            * nav_rt or brdc_rt as archive center allows to download *real time* nav files from the BKG server
            * rgp (IGN's RGP main server at St Mandé)
            * rgp_ensg (IGN's RGP, secondary server at ENSG, Marne-la-Vallée)
        _ not reimplemented yet _:
            * rgp_1Hz (IGN's RGP, all the 24 hourly rinex for the day will be downloaded)
            * renag
            * ovsg
            * unavco
            * geoaus (Geosciences Australia)
            * ens_fr

    output_dir : str
        the root directory on your local drive were to store the RINEXs

    archtype : str
        string describing how the archive directory is structured, e.g:
            * stat
            * stat/year
            * stat/year/doy
            * year/doy
            * year/stat
            * week/dow/stat
            * etc ...

    user & passwd : str
        user & password for a secure server

    path_ftp_crawled_files_save : str
        will save at the given path (directory+filename) in a CSV file containing
        the list of the existing RINEXs found on the server by the FTP crawler.
        It allows to use this list directly if one face a timeout during
        the download part.

    path_ftp_crawled_files_load : str
        load and use the list of the existing RINEXs found on the FTP server,
        generated by a previous run of the FTP crawler (called by
        download_gnss_rinex or directly by crawl_ftp_files).
        a new call of crawl_ftp_files can be bypassed with
        `skip_crawl`.

    skip_crawl : bool
        when a list of the existing RINEXs found on the FTP server
        is provided with path_ftp_crawled_files_load, skip
        a new call of crawl_ftp_files.
        Then, we assume the provided list as a complete one,
        ready for the download step

    path_all_ftp_files_save : str
        will save at the given path (directory+filename) in a CSV file
        ALL the remote files found on the FTP server.

    quiet_mode : bool
        List the available RINEXs without downloading them.
        Useful only if path_ftp_crawled_files_save is given

    final_archive_for_sup_check : str
        The final archive path or a file containing the archived RINEXs in
        their final destination.
        useful if the final archive is different from archive_dir
        *** not re-implemented yet ***

    force : bool
        Force the download even if the file already exists locally

    no_rnx2 & no_rnx3 : bool
        limit the search/download to RINEX2 (short names) and/or
        RINEX3 (long names) depending on the boolean given

    Returns
    -------
    out_tup_lis : List of tuples
        Returns a list of tuples containing
        the local path of the downloaded file and
        a boolean indicating whether the download was successful.
        e.g. [(local_path1, True), (local_path2, False), ...]

    Minimal exemple
    ---------------
        >>> statdic = dict()
        >>> statdic['igs_cddis'] = ['ZIMM','tlse']
        >>> archive_dir = '/home/USER/test_dl_rnx'
        >>> startdate = dt.datetime(2020,1,1)
        >>> enddate = dt.datetime(2020,1,31)
        >>> geodezyx.operational.download_gnss_rinex(statdic, output_dir, startdate, enddate)
    """

    date_range = conv.dt_range(startdate, enddate)

    log.info("dates: %s to %s", startdate, enddate)

    if path_ftp_crawled_files_load:
        table = pd.read_csv(path_ftp_crawled_files_load)
    else:
        table = gen_crawl_table(
            statdico, date_range, output_dir, archtype, no_rnx2, no_rnx3
        )

    if len(table) == 0:
        log.error("No RINEX files found for the given criteria.")
        return None

    if skip_crawl:
        table_crawl = table
        files_all = pd.Series([], dtype=str)
        files_loc = pd.Series([], dtype=str)
    else:
        table_crawl, files_all, files_loc = crawl_ftp_files(
            table,
            sftp="auto",
            user=user,
            passwd=passwd,
            path_ftp_crawled_files_save=path_ftp_crawled_files_save,
            path_all_ftp_files_save=path_all_ftp_files_save,
            force=force,
        )

    #### get only the valid (true) url
    table_dl = table_crawl.loc[table_crawl["url_true"].dropna().index]

    # Initialize output list
    out_tup_lis = []

    # Only download if files are available and not in quiet mode
    if len(table_dl) == 0:
        log.error(
            "no valid RINEX URL found/selected on the FTP server, check your inputs"
        )
    elif quiet_mode:
        log.warning("quiet mode, no download was performed")
    else:
        out_tup_lis = dlutils.ftp_downld_front(
            table_dl["url_true"].values,
            table_dl["outdir"].values,
            parallel_download=parallel_download,
            secure_ftp=table_dl["sftp"].values,
            user=user,
            passwd=passwd,
            force=force,
        )

    ### add the local paths to the output tuples
    if len(files_loc) > 0:
        loc_tup_lis = [(f, True) for f in files_loc]
        out_tup_lis_fin = loc_tup_lis + out_tup_lis
    else:
        out_tup_lis_fin = out_tup_lis

    log.info(
        "RINEX files fetched: total: %d, downloaded: %d, already here: %d",
        len(out_tup_lis_fin),
        len(out_tup_lis),
        len(files_loc),
    )

    return out_tup_lis


def gen_crawl_table(statdico, date_range, output_dir, archtype, no_rnx2, no_rnx3):
    """
    Generate a crawl table for RINEX file downloads.

    Parameters
    ----------
    statdico : dict
        Dictionary mapping data centers to station lists.
    date_range : list
        List of datetime objects for the date range.
    output_dir : str
        Root output directory.
    archtype : str
        Archive directory structure type.
    no_rnx2 : bool
        Skip RINEX2 files if True.
    no_rnx3 : bool
        Skip RINEX3 files if True.

    Returns
    -------
    pd.DataFrame
        Table with download metadata for each RINEX file.
    """
    table_proto = []

    for datacenter, site_lis in statdico.items():
        log.info("datacenter/stations: %s/%s", datacenter, " ".join(site_lis))

        for date, site in itertools.product(date_range, site_lis):
            urldic, sftp, _ = _server_select(datacenter, site, date)
            if not urldic:
                continue

            outdir = effective_save_dir(output_dir, site, date, archtype)

            for rnxver, rnxurl in urldic.items():
                if (rnxver == 2 and no_rnx2) or (rnxver == 3 and no_rnx3):
                    continue
                table_proto.append((date, site, outdir, rnxver, rnxurl, sftp))

    # Create DataFrame with all collected data
    table = pd.DataFrame(
        table_proto, columns=["date", "site", "outdir", "ver", "url_theo", "sftp"]
    )

    # Add status columns
    table["crawled"] = False
    table["ok_dwl"] = False
    table["ok_loc"] = False
    table["url_true"] = None
    table["rnxnam"] = ""

    # Parse URL components
    urlpaths = table["url_theo"].apply(pathlib.Path)
    table["rnxrgx"] = urlpaths.apply(lambda p: p.name)
    table["host"] = urlpaths.apply(lambda p: p.parts[1])
    table["dir"] = urlpaths.apply(lambda p: os.path.join(*p.parts[2:-1]))

    return table



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 26/02/2025 11:01:41

@author: psakic
"""
from .legacy import  *
from .gins_common  import *
from .gins_prairie import *
from .gins_orbclk_concat  import *
from .gins_dirs_gen import *
from .gins_dirs_run import *
from .spotgins_run import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 27/02/2025 10:37:26

@author: psakic
"""

import argparse
import getpass
import os
import subprocess
import datetime as dt
import re
import sys

from geodezyx import conv, utils
import geodezyx.operational.gins_runner.gins_common as gynscmn

#### Import the logger
import logging

log = logging.getLogger("geodezyx")


def read_credentials(file_path):
    with open(file_path, "r") as file:
        login = file.readline().strip()
        password = file.readline().strip()
    return login, password


def download_rsync(
    file_list,
    remote_user,
    remote_host,
    remote_path,
    local_destination,
    rsync_options=None,
    password=None,
):
    """
    Downloads a list of files using rsync.
    """
    if not rsync_options:
        rsync_options = [
            "--archive",
            "--verbose",
            "--compress",
            "--progress",
            "--relative",
            "--copy-links",
            "--no-perms",
            "--no-owner",
            "--no-group",
            "--omit-dir-times",
        ]  # Default options: archive mode, verbose, and compression

    # Construct the remote source path
    remote_source = f"{remote_user}@{remote_host}:{remote_path}"

    if password:
        rsync_base = ["sshpass", "-p", password, "rsync"] + rsync_options
    else:
        rsync_base = ["rsync"] + rsync_options

    # if exclude_compress:
    #   '--ignore-existing' '--exclude-from=<(echo "$(find /destination -name '*.gz' | sed 's/\.gz$//')"')'

    tmp_rsync_file_lis = f"/tmp/" + utils.get_timestamp() + "_tmp_rsync_file_list.lst"
    utils.write_in_file("\n".join(file_list), tmp_rsync_file_lis)

    # Construct the rsync command
    # /./ : https://askubuntu.com/questions/552120/preserve-directory-tree-while-copying-with-rsync
    rsync_cmd = (
        rsync_base
        + ["--files-from", tmp_rsync_file_lis]
        + [f"{remote_source}/./", f"{local_destination}/"]
    )

    log.debug("Rsync command: %s", " ".join(rsync_cmd))

    # Run the rsync command
    process = subprocess.run(rsync_cmd, stderr=sys.stderr, stdout=sys.stdout, text=True)

    if process.returncode != 0:
        log.error("Rsync failed with return code :( %d", process.returncode)
    else:
        log.info("Rsync completed successfully :)")

    os.remove(tmp_rsync_file_lis)
    return


def bdgins_update(
    date_srt=dt.datetime(2020, 5, 3),
    date_end=None,
    dir_bdgins="",
    login="",
    password="",
    compress=False,
):
    """
    Update the BDGINS repository with the necessary files
    for the given date range.

    Parameters
    ----------
    date_srt : datetime
        The start date for the update.
        The default is 2020-05-03 (origin of G20 products, end of GPS's Selectable Avaiability).
    date_end : datetime
        The end date for the update.
        If not provided, the default is the 15 days ago.
    dir_bdgins : str, optional
        The directory for BDGINS. Defaults to an empty string.
    login : str, optional
        The login for the remote server. Defaults to an empty string.
    password : str, optional
        The password for the remote server. Defaults to an empty string.
    compress : bool, optional
        Whether to gzip-compress the clock files.
        experimental, and not recommended
        Defaults to False.
    
    Returns
    -------
    None
    """
    if not date_end:
        date_end = dt.datetime.now() - dt.timedelta(days=15)
        date_end = conv.round_dt(date_end, "1D", mode="floor", python_dt_out=False)
        date_end = conv.numpy_dt2dt(date_end)

    if not dir_bdgins:
        dir_bdgins = os.path.join(gynscmn.get_gin_path(True), "data")

    if not login:
        home = os.path.expanduser("~")
        ginspcrc = os.path.join(home, ".ginspc")
        if os.path.isfile(ginspcrc):
            login = open(ginspcrc, "r").read().split("=")[1].strip()
        else:
            login = getpass.getuser()

    date = date_srt

    ###### LIST INITIALISATION
    ### misc files
    list_misc = []

    ### full folders
    list_antex = []
    list_constell = []
    list_exe_ppp = []
    list_lunisolaires = []
    list_macromod = []
    list_maree_polaire = []
    list_pole = []
    list_prairie = []

    ### time dependant files
    list_tropo_vmf1 = []
    list_iono_igs = []
    list_orbite_mg3 = []

    list_orbite_g20 = []
    list_orbex_g20 = []
    list_horl_g20 = []

    ###### LIST FILL
    ### full folders
    # (folder's path is added in the rsync command, with subdir destination variable
    ## list_antex.extend(["/"]) ## temporary off because updated continuously
    ## list_constell.extend(["/"])
    ## list_exe_ppp.extend(["/"])
    ## list_lunisolaires.extend(["/"]) ## temporary off because updated continuously
    ## list_macromod.extend(["/"])
    ## list_maree_polaire.extend(["/"])
    ## list_pole.extend(["/"])
    list_prairie.extend(["/"])

    ### misc files: the needed files are considered individually
    # some of them are a redundancy since they must be downloaded in the full folders
    list_misc.extend([f"prairie/igs_satellite_metadata.snx"])
    list_misc.extend([f"pole/nominal_NRO"])
    list_misc.extend([f"ANTEX/igs20.atx"])
    list_misc.extend([f"EXE_PPP/valap_static"])
    list_misc.extend([f"macromod/gnss.xml"])
    list_misc.extend([f"lunisolaires/de440bdlf.ad"])
    list_misc.extend([f"maree_polaire/loading/nominal"])
    l_fil_cons = [
        "constellation_gps.infos",
        "histocom.infos",
        "histogal.infos",
        "historik_glonass",
        "igs_satellite_metadata.snx",
    ]
    list_misc.extend(["constell/" + f for f in l_fil_cons])

    # here is an hybrid tropo/misc file
    list_tropo_vmf1.append(f"orography_ell")

    ### time dependant files
    while date <= date_end + dt.timedelta(days=2): # 2 days later is need for cat orb/clk
        day = str(date.day).zfill(2)
        month = str(date.month).zfill(2)
        year = date.year
        yy = str(year)[2:]

        wk, wkday = conv.dt2gpstime(date)
        doy = str(conv.dt2doy(date)).zfill(3)

        list_tropo_vmf1.extend(
            [
                f"{year}/VMFG_{year}{month}{day}.H00",
                f"{year}/VMFG_{year}{month}{day}.H06",
                f"{year}/VMFG_{year}{month}{day}.H12",
                f"{year}/VMFG_{year}{month}{day}.H18",
            ]
        )
        list_iono_igs.append(f"{year}/igsg{doy}0.{yy}i.Z")
        list_orbite_mg3.append(f"mg3{wk}{wkday}.sp3.Ci9PAU")
        list_orbite_g20.append(f"G20{wk}{wkday}.gin")
        list_orbex_g20.append(f"G20{wk}{wkday}.obx.gz")
        list_horl_g20.append(f"hogps_g20{wk}{wkday}")
        date += dt.timedelta(days=1)

    ###### DESTINATION FOLDERS
    dest_subdir_dic = {
        ### full folders
        #"ANTEX": list_antex,
        #"constell": list_constell,
        #"EXE_PPP": list_exe_ppp,
        #"lunisolaires": list_lunisolaires,
        #"macromod": list_macromod,
        #"maree_polaire": list_maree_polaire,
        "prairie": list_prairie,
        ### misc files
        ".": list_misc,  ## for misc files, destination is in the input path (.)
        ### time dependant files
        "tropo_vmf1": list_tropo_vmf1,
        "ionosphere/igs": list_iono_igs,
        "orbites/SP3/re3": list_orbite_mg3,
        "mesures/gps/orbites/G20": list_orbite_g20,
        "mesures/gps/orbex/G20": list_orbex_g20,
        "mesures/gps/horloges30/G20": list_horl_g20,
    }

    create_dir(dir_bdgins, subdirs=dest_subdir_dic.keys())
    os.chdir(dir_bdgins)

    for subdir, files_list in dest_subdir_dic.items():
        subdir_print = "misc. files" if subdir == "." else subdir
        log.info("Downloading files for %s", subdir_print)

        local_subdir_fullpath = os.path.join(dir_bdgins, subdir)
        remote_subdir_fullpath = os.path.join("/home/gins/MIROIR_STAF", subdir)

        download_rsync(
            files_list,
            login,
            "tite.get.obs-mip.fr",
            remote_subdir_fullpath,
            local_subdir_fullpath,
            password,
        )

    # compress the clock files (experimental, and not recommended)
    if compress:
        for dirr in ["mesures/gps/horloges30/G20", "mesures/gps/orbites/G20"]:
            d = os.path.join(dir_bdgins, dirr)
            cmd = [
                "find",
                d,
                "-name",
                "!",
                ".gz",
                "-exec",
                "gzip",
                "-v",
                "{}",
                r"\;",
            ]
            process = subprocess.run(
                cmd, stderr=sys.stderr, stdout=sys.stdout, text=True
            )

    return None


def create_dir(parent_dir, subdirs):
    if not os.path.exists(parent_dir):
        os.makedirs(parent_dir)

    for subdir in subdirs:
        subdir_fullpath = os.path.join(parent_dir, subdir)
        if not os.path.exists(subdir_fullpath):
            os.makedirs(subdir_fullpath)


def main():
    parser = argparse.ArgumentParser(description="Update BDGINS repository.")
    parser.add_argument(
        "-s",
        "--date_srt",
        type=lambda s: conv.date_pattern_2_dt(s),
        required=False,
        default=dt.datetime(2020, 5, 3),
        help=(
            "Start date for the update in various formats. "
            "Default is 2020-05-03 "
            "(origin of G20 products, end of GPS's Selective Availability)."
        ),
    )
    parser.add_argument(
        "-e",
        "--date_end",
        type=lambda s: conv.date_pattern_2_dt(s),
        required=False,
        default=None,
        help=(
            "End date for the update in various formats. "
            "Default is 15 days before the current date."
        ),
    )
    parser.add_argument(
        "-d",
        "--dir_bdgins",
        help=(
            "Directory for the BDGINS repository. "
            "Defaults to a predefined path if not provided."
        ),
        required=False,
    )
    parser.add_argument(
        "-l",
        "--login",
        help=(
            "Login for the remote server. "
            "Defaults to the current user or a value from the `.ginspc` file."
        ),
        required=False,
    )
    parser.add_argument(
        "-p",
        "--password",
        help=(
            "Password for the remote server. "
            "Defaults to an empty string if not provided."
        ),
        required=False,
        default=None,
    )
    parser.add_argument(
        "-c",
        "--compress",
        help=(
            "Flag to enable gzip compression of clock files. "
            "Defaults to `False` and is experimental."
        ),
        action="store_true",
        default=False,
    )
    args = parser.parse_args()

    bdgins_update(
        date_srt=args.date_srt,
        date_end=args.date_end,
        login=args.login,
        password=args.password,
        dir_bdgins=args.dir_bdgins,
        compress=args.compress,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.operational contains functions to run the 
time series velocities estimation software MIDAS. 

it can be imported directly with:
from geodezyx import operational

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import glob
#### Import the logger
import logging
import os
import shutil
import subprocess

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

#### geodeZYX modules
from geodezyx import conv
from geodezyx import stats
from geodezyx import utils

log = logging.getLogger('geodezyx')

##########  END IMPORT  ##########


#
#  __  __ _____ _____           _____
# |  \/  |_   _|  __ \   /\    / ____|
# | \  / | | | | |  | | /  \  | (___
# | |\/| | | | | |  | |/ /\ \  \___ \
# | |  | |_| |_| |__| / ____ \ ____) |
# |_|  |_|_____|_____/_/    \_\_____/
#
#


def midas_run(
    tenu_file_path,
    work_dir="",
    path_midas_soft="",
    step_file_path="",
    with_plot=True,
    keep_plot_open=True,
):
    ### Prepare paths of facultative arguments
    if not work_dir:
        work_dir = os.path.dirname(tenu_file_path)
    if not path_midas_soft:
        path_midas_soft = "midas.e"

    ### SECURITY : rm all previous MIDAS files
    MIDAS_files_rm = [
        "MIDAS.STEPIN",
        "MIDAS.STEPOUT",
        "MIDAS.ERR",
        "MIDAS.TENV",
        "MIDAS.RENV",
        "MIDAS.TENU",
        "MIDAS.RENU",
        "MIDAS.VEL",
        "fort.91",
    ]

    for fil in MIDAS_files_rm:
        fil_full_path = os.path.join(work_dir, fil)
        if os.path.isfile(fil_full_path):
            log.info("old %s removed", fil)
            os.remove(fil_full_path)

    ### Find the Station Name
    DF = pd.read_table(tenu_file_path, comment="*", header=-1, delim_whitespace=True)
    stat = list(set(DF[0]))[0]

    ### Prepare the tmp input
    ### The file is called MIDAS.TENU in any case
    os.chdir(work_dir)
    work_file_path_tenu = work_dir + "/MIDAS.TENU"
    shutil.copy(tenu_file_path, work_file_path_tenu)

    ### Prepare the tmp STEP FILE input
    step_OK = False
    if step_file_path and not utils.empty_file_check(step_file_path):
        work_file_path_step = work_dir + "/MIDAS.STEP"
        shutil.copy(step_file_path, work_file_path_step)
        step_OK = True
    else:
        work_file_path_step = ""

    ### Prepare the name of the outputed file
    work_file_path_vel = os.path.join(work_dir, "MIDAS.VEL")

    ### LETS RUN !
    command = path_midas_soft
    log.info("LAUNCHING : %s for stat. %s", command, stat)
    p = subprocess.Popen(
        "",
        executable="/bin/bash",
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = p.communicate(command.encode())
    # logs files
    vel_file = os.path.join(work_dir, "MIDAS.VEL")
    vel_file_obj = open(vel_file, "w")
    vel_file_obj.write(stdout.decode("utf-8"))
    vel_file_obj.close()

    exec_OK = False

    if not utils.empty_file_check(work_file_path_vel):
        exec_OK = True
        log.info("MIDAS.VEL exists for %s, should be OK :)", stat)

    if not utils.empty_file_check(os.path.join(work_dir, "MIDAS.ERR")):
        log.error("MIDAS.ERR is not empty for, must be checked :(", stat)

    if stderr:
        log.warning("err.log is not empty, must be checked !")
        log.warning(stderr.decode("utf-8"))
        log.warning("")
        err_file = os.path.join(work_dir, stat + ".err.log")
        err_file_obj = open(err_file, "w")
        err_file_obj.write(stderr.decode("utf-8"))
        err_file_obj.close()

    #### PLOT
    if exec_OK and with_plot:
        fig = midas_plot(work_file_path_tenu, work_file_path_vel, work_file_path_step)
        work_file_path_plot = os.path.join(work_dir, "plot." + stat.upper())
        for ext in (".pdf", ".png"):
            plt.savefig(work_file_path_plot + ext)
        if not keep_plot_open:
            plt.close(fig)

    #### RENAME ALL MIDAS FILES WITH A STATION PREFIX
    for fname in glob.glob("*MIDAS*"):
        os.rename(fname, fname.lower() + "_" + stat.upper())

    return None


def midas_vel_files_2_pandas_DF(vel_files_in):
    """
    Convert MIDAS Velocity files to a Pandas DataFrame

    Parameters
    ----------
    vel_files_in : str or list of str
        if list of str, will consider directly the files inside the list
        if str, can be the path of a single file, or a generic (wilcard) path
        to consider several files
    Returns
    -------
    DF : Pandas DataFrame
    """

    if utils.is_iterable(vel_files_in):
        L_vel_files = vel_files_in
    else:
        L_vel_files = glob.glob(vel_files_in)

    if len(L_vel_files) > 1:
        work_dir = os.path.dirname(L_vel_files[0])
        vel_file_opera = utils.cat(work_dir + "/MERGED_VEL", *L_vel_files)
    else:
        vel_file_opera = L_vel_files[0]

    DF = pd.read_table(vel_file_opera, comment="*", header=-1, delim_whitespace=True)

    DF.columns = [
        "Station",
        "soft",
        "epoch_first",
        "epoch_last",
        "duration",
        "nb_epoch_all",
        "nb_epoch_good",
        "nb_pairs",
        "V_East",
        "V_North",
        "V_Up",
        "sV_East",
        "sV_North",
        "sV_Up",
        "offset_e_1st_epoch",
        "offset_n_1st_epoch",
        "offset_u_1st_epoch",
        "outlier_ratio_e",
        "outlier_ratio_n",
        "outlier_ratio_u",
        "std_velo_pair_e",
        "std_velo_pair_n",
        "std_velo_pair_u",
        "nb_steps",
    ]

    return DF


def midas_plot(path_tenu, path_vel="", path_step=""):
    """
    based on a plot for TimeSeriePoint
    """
    import geoclass as gcls

    DF = pd.read_table(path_tenu, header=-1, delim_whitespace=True)
    stat = list(set(DF[0]))[0]

    if path_vel:
        DFvel = midas_vel_files_2_pandas_DF(path_vel)
        plot_vel = True
    else:
        plot_vel = False

    if path_step:
        DFstep = pd.read_table(path_step, header=-1, delim_whitespace=True)
        Discont = conv.year_decimal2dt(list(DFstep[1]))

    T = DF[1]
    Tdt = conv.year_decimal2dt(T)
    E = DF[2]
    N = DF[3]
    U = DF[4]

    TS = gcls.TimeSeriePoint()

    TS.from_list(Tdt, E, N, U, "ENU")
    TS.meta_set(path_tenu, stat, stat)

    if path_step:
        TS.set_discont(Discont)
        log.info("Discontinuity", Discont)

    #### PLOT ########
    TSplot_fig = TS.plot(fig=plt.figure())
    if path_step:
        TS.plot_discont(fig=TSplot_fig)
    ##################
    TSplot_axe = TSplot_fig.axes

    a_keys_list = ["V_East", "V_North", "V_Up"]

    b_keys_list = ["offset_e_1st_epoch", "offset_n_1st_epoch", "offset_u_1st_epoch"]

    sigma_key_list = ["sV_East", "sV_North", "sV_Up"]

    if plot_vel:
        for i, (a_key, b_key, sigma_key) in enumerate(
            zip(a_keys_list, b_keys_list, sigma_key_list)
        ):
            ax = TSplot_axe[i]

            a = DFvel[a_key][0]
            b = DFvel[b_key][0]
            sigma = DFvel[sigma_key][0]

            _, Vlin = stats.linear_reg_getvalue(T - T[0], a, b)

            ax.plot(Tdt, Vlin, c="xkcd:dark orange")
            text_vel = (
                str(np.round(a * 1000, 4))
                + " +/- "
                + str(np.round(sigma * 1000, 4))
                + " mm/yr"
            )

            ax.text(
                0.8,
                0.1,
                text_vel,
                horizontalalignment="center",
                verticalalignment="center",
                transform=ax.transAxes,
            )

    return TSplot_fig



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.reffram contains functions related to
 Kepler's element operations.

it can be imported directly with:
from geodezyx import reffram

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

import datetime as dt

#### Import the logger
import logging

########## BEGIN IMPORT ##########
#### External modules
import numpy as np
import pandas as pd

# from pytwobodyorbit import TwoBodyOrbit

#### geodeZYX modules
from geodezyx import conv

log = logging.getLogger("geodezyx")

##########  END IMPORT  ##########


def eci_2_kepler_elts(pos, vel, rad2deg=True, mu=3.9860044188e14):
    """
    Convert ECI coordinates > Kepler's elements

    Parameters
    ----------
    pos : array
        Position in m.
        Can be a 3-element (x,y,z) array (simple point)
        or a (N,3)-shaped numpy array
    vel : array
        Velocity in m/s.
        Can be a 3-element (vx,vy,vz) array (simple point)
        or a (N,3)-shaped numpy array
    rad2deg : bool, optional
        convert Keplerian's angles (anomalies) to deg. The default is True.
    mu : float, optional
        Standard gravitational parameter. The default is 3.9860044188e14.

    Returns
    -------
    a,ecc,i,o_peri,o_lan,m: floats
        Kepler's elements:

        * a : semi-major axis
        * ecc : orbit eccentricity
        * i : orbit inclination
        * o_peri : argument of periapsis ω
        * o_lan : longitude of the ascending node Ω
        * m : mean anomaly m

    Note
    ----
    Be sure your input is in ECI (SP3 are in ECEF for instance).

    If neeeded, use ``conv.ecef2eci()`` (gross results)
    or ``reffram.orb_df_crf2trf(inv_trf2crf=True)`` (precise results)

    You can get the velocity with ``reffram.orb_df_velocity_calc()``

    Source
    ------
    https://downloads.rene-schwarz.com/download/M002-Cartesian_State_Vectors_to_Keplerian_Orbit_Elements.pdf

    """
    p = np.array(pos)
    v = np.array(vel)

    #### case for one point only
    if len(p.shape) < 2:
        p = np.expand_dims(p, axis=-1).T
    if len(v.shape) < 2:
        v = np.expand_dims(v, axis=-1).T

    pnorm = np.linalg.norm(p, axis=1)
    vnorm = np.linalg.norm(v, axis=1)

    ### orbital momentum vector h
    h = np.cross(p, v)
    hnorm = np.linalg.norm(h, axis=1)

    ### eccentricity vector eccvec & orbit eccentricity ecc
    eccvec = (np.cross(v, h) / mu) - (p / np.expand_dims(pnorm, axis=-1))
    ecc = np.linalg.norm(eccvec, axis=1)

    ### true anomaly ν (nu)
    n = np.cross(np.array([0, 0, 1]), h)
    nnorm = np.linalg.norm(n, axis=1)

    dot_pv_sup0 = np.einsum("ij,ij->i", p, v) >= 0
    nu_sup0 = np.arccos(np.einsum("ij,ij->i", eccvec, p) / (ecc * pnorm))
    nu_inf0 = 2 * np.pi - nu_sup0

    nu = np.zeros(len(dot_pv_sup0))
    nu = nu + nu_sup0 * dot_pv_sup0
    nu = nu + nu_inf0 * np.logical_not(dot_pv_sup0)

    ### orbit inclination i
    i = np.arccos(h[:, 2] / hnorm)

    ### eccentric anomaly E (called e here)
    # e1 = 2*np.arctan(np.tan(nu/2)/np.sqrt((1+ecc)/(1-ecc)))
    e2 = 2 * np.arctan2(np.tan(nu / 2), np.sqrt((1 + ecc) / (1 - ecc)))
    e = e2

    ### longitude of the ascending node Ω (o_lan)
    n1_sup0 = n[:, 1] >= 0

    omega_lan_sup0 = np.arccos(n[:, 0] / nnorm)
    omega_lan_inf0 = 2 * np.pi - omega_lan_sup0

    o_lan = np.zeros(len(n1_sup0))
    o_lan = o_lan + omega_lan_sup0 * n1_sup0
    o_lan = o_lan + omega_lan_inf0 * np.logical_not(n1_sup0)

    ### argument of periapsis ω (o_peri)
    eccvec2_sup0 = eccvec[:, 2] >= 0
    o_peri_sup0 = np.arccos((np.einsum("ij,ij->i", n, eccvec)) / (ecc * nnorm))
    o_peri_inf0 = 2 * np.pi - o_peri_sup0

    o_peri = np.zeros(len(eccvec2_sup0))
    o_peri = o_peri + o_peri_sup0 * eccvec2_sup0
    o_peri = o_peri + o_peri_inf0 * np.logical_not(eccvec2_sup0)

    ### mean anomaly m, (Kepler’s Equation)
    m = e - ecc * np.sin(e)

    ### semi-major axis a
    a = ((2 / pnorm) - (vnorm**2 / mu)) ** -1

    #### case for one point only (bring it back to a scalar)
    sqzflt = lambda x: np.float64(x.squeeze())

    a = sqzflt(a)
    ecc = sqzflt(ecc)
    i = sqzflt(i)
    o_lan = sqzflt(o_lan)
    o_peri = sqzflt(o_peri)
    m = sqzflt(m)

    if rad2deg:
        i = np.rad2deg(i)
        o_peri = np.rad2deg(o_peri)
        o_lan = np.rad2deg(o_lan)
        m = np.rad2deg(m)

    #### o_peri and m seems to be inverted, must be investigated!!!
    return a, ecc, i, o_peri, o_lan, m


def eci_2_kepler_elts_mono(p, v, rad2deg=True, mu=3.9860044188e14):
    """
    **Kept for debug purposes, use** ``eci_2_kepler_elts(pos, vel)`` **instead**

    Convert ECI coordinates > Kepler's elements

    Parameters
    ----------
    p : array
        Position.
    v : array
        Velocity.
    rad2deg : bool, optional
        canvert Keplerian's angles (anomalies) to deg. The default is True.
    mu : float, optional
        Standard gravitational parameter. The default is 3.9860044188e14.

    Returns
    -------
    a,ecc,i,omega_periarg,omega_lan,m : floats
        Kepler's elements.
    """
    log.warning(
        "use this function for debug purposes only, use eci_2_kepler_elts instead"
    )
    pnorm = np.linalg.norm(p)
    vnorm = np.linalg.norm(v)

    ### orbital momentum vector h
    h = np.cross(p, v)
    hnorm = np.linalg.norm(h)

    ### eccentricity vector ecc
    ecc = (np.cross(v, h) / mu) - (p / pnorm)
    ecc_norm = np.linalg.norm(ecc)

    ### true anomaly ν (nu)
    n = np.cross(np.array([0, 0, 1]), h)
    nnorm = np.linalg.norm(n)
    if np.dot(p, v) >= 0:
        nu = np.arccos(np.dot(ecc, p) / (ecc_norm * pnorm))
    else:
        nu = 2 * np.pi - np.arccos(np.dot(ecc, p) / (ecc_norm * pnorm))

    ### orbit inclination i
    i = np.arccos(h[2] / hnorm)

    ### orbit eccentricity ecc
    ecc = np.linalg.norm(ecc)

    ### eccentric anomaly E (called e here)
    # e1 = 2*np.arctan(np.tan(nu/2)/np.sqrt((1+ecc)/(1-ecc)))
    e2 = 2 * np.arctan2(np.tan(nu / 2), np.sqrt((1 + ecc) / (1 - ecc)))
    e = e2

    ### longitude of the ascending node Ω (omega_lan)
    if n[1] >= 0:
        omega_lan = np.arccos(n[0] / nnorm)
    else:
        omega_lan = 2 * np.pi - np.arccos(n[0] / nnorm)

    ### argument of periapsis ω (omega_periarg)
    if ecc[2] >= 0:
        omega_periarg = np.arccos((np.dot(n, ecc)) / (nnorm * ecc_norm))
    else:
        omega_periarg = 2 * np.pi - np.arccos((np.dot(n, ecc)) / (nnorm * ecc_norm))

    ### mean anomaly m, (Kepler’s Equation)
    m = e - ecc * np.sin(e)

    ### semi-major axis a
    a = ((2 / pnorm) - (vnorm**2 / mu)) ** -1

    if rad2deg:
        i = np.rad2deg(i)
        omega_periarg = np.rad2deg(omega_periarg)
        omega_lan = np.rad2deg(omega_lan)
        m = np.rad2deg(m)

    return a, ecc, i, omega_periarg, omega_lan, m


def extrapolate_orbit_kepler(p, v, t, t0, mu=3.9860044188e14):
    from pytwobodyorbit import TwoBodyOrbit

    orbit = TwoBodyOrbit("", mu=mu)  # create an instance
    orbit.setOrbCart(t0, p, v)  # define the orbit
    pout, vout = orbit.posvelatt(t)  # get position and velocity at t1
    kepl = orbit.elmKepl()  # get classical orbital elements

    return pout, vout, kepl


def extrapolate_sp3_data_frame(
    df_sp3,
    step=900,
    n_step=9,
    backward=True,
    forward=True,
    until_backward=None,
    until_forward=None,
    return_all=True,
):
    """
    Extrapolate the positions in a SP3 based on the first/last
    epochs' position

    Parameters
    ----------
    df_sp3 : DataFrame
        Input Orbit DataFrame (i.e. generated by files_rw.read_sp3).
    step : int, optional
        step between two epochs. The default is 900.
    n_step : int, optional
        number of epochs to interpolate. The default is 9.
    backward and forward : bool, optional
        extrapolate for the day before/after. The default is True.
    return_all : bool, optional
        if True, returns the input DF enhanced with the extrapolated values
        if False, returns only the extrapolated values.
        The default is True.
    until_backward & until_backward : datetime, optional
        epoch until then the extrapolation has to be done.
        Override n_step

    Returns
    -------
    df_orb_out : DataFrame
        Orbit DataFrame with extrapolated values
        (see return_all).

    """
    new_epoch_stk = []

    for sat in df_sp3["sat"].unique():
        log.info("extrapolate: %s", sat)
        df_sat = df_sp3[(df_sp3["sat"] == sat) & (df_sp3["type"] == "p")].copy()
        df_sat.sort_values("epoch", inplace=True)

        df_line_dummy = df_sat.iloc[0].copy()
        df_line_dummy["clk"] = 999999.999999

        xyz = df_sat[["x", "y", "z"]].values * 1000
        tgps = conv.numpy_dt2dt(df_sat["epoch"].values)
        tutc = conv.dt_gpstime2dt_utc(tgps)
        tsec = [e.total_seconds() for e in (tutc - tutc[0])]

        p = conv.ecef2eci(xyz, tutc)
        v = np.gradient(p, tsec, axis=0)

        def extrapo_intern_fct(i_ref, coef, until):
            ##
            ## backward :
            ## i_ref , coef = 0,-1
            ##
            ## forward :
            ## i_ref , coef = -1,1
            ##
            ## CORRDS ARE GIVEN IN ECI HERE !!!
            ##
            from pytwobodyorbit import TwoBodyOrbit

            orbit_back = TwoBodyOrbit("orbit", mu=3.9860044188e14)  # create an instance
            orbit_back.setOrbCart(tsec[i_ref], p[i_ref], v[i_ref])  # define the orbit

            if until:
                t_rang_strt, t_rang_end = list(sorted([tutc[i_ref], until]))
                range = conv.dt_range(t_rang_strt, t_rang_end, 0, step)
                n_step_intern = len(range) - 1
            else:
                n_step_intern = n_step

            for t in np.arange(step, n_step_intern * step + 1, step):
                pout, vout = orbit_back.posvelatt(coef * t)
                epoc = tgps[i_ref] + coef * dt.timedelta(seconds=int(t))
                df_line_new = df_line_dummy.copy()
                df_line_new[["x", "y", "z"]] = pout
                df_line_new["epoch"] = pd.Timestamp(epoc)

                new_epoch_stk.append(df_line_new)

            return pout, vout

        ### backward
        if backward:
            extrapo_intern_fct(0, -1, until_backward)

        ### forward
        if forward:
            extrapo_intern_fct(-1, 1, until_forward)

    df_new_epoch = pd.DataFrame(new_epoch_stk)

    tutc_new_epoch = conv.dt_gpstime2dt_utc(
        conv.numpy_dt2dt(df_new_epoch["epoch"].values)
    )
    df_new_epoch[["x", "y", "z"]] = conv.eci2ecef(
        df_new_epoch[["x", "y", "z"]].values, tutc_new_epoch
    )
    df_new_epoch[["x", "y", "z"]] = df_new_epoch[["x", "y", "z"]] * 10**-3

    if return_all:
        df_orb_out = pd.concat((df_sp3, df_new_epoch))
    else:
        df_orb_out = df_new_epoch

    df_orb_out.reset_index(drop=True, inplace=True)
    df_orb_out.sort_values(["sat", "epoch"], inplace=True)
    df_orb_out.reset_index(drop=True, inplace=True)

    return df_orb_out
